# -*- coding: utf-8 -*-
""" Perform tests on methods to create testsignals."""
import matplotlib
import numpy as np
from numpy.testing import assert_almost_equal
from pytest import approx

from examples.working_with_signals import demonstrate_signal
from PyDynamic.misc.testsignals import (
    GaussianPulse,
    rect,
    shocklikeGaussian,
    sine,
    squarepulse,
)

N = 2048
Ts = 0.01


def get_timestamps(t_0=0, t_n=N * Ts, d_t=Ts):
    # Compute equally spaced timestamps between t_0 and t_n.
    return np.arange(t_0, t_n, d_t)


time = get_timestamps(0, N * Ts, Ts)
t0 = N / 2 * Ts


def test_shocklikeGaussian():
    m0 = 1 + np.random.rand() * 0.2
    sigma = 50 * Ts
    # zero noise
    x = shocklikeGaussian(time, t0, m0, sigma, noise=0.0)
    assert_almost_equal(np.max(x), m0)
    assert np.std(x[: N // 10]) < 1e-10
    # noisy signal
    nstd = 1e-2
    x = shocklikeGaussian(time, t0, m0, sigma, noise=nstd)
    assert_almost_equal(np.round(np.std(x[: N // 10]) * 100) / 100, nstd)


def test_GaussianPulse():
    m0 = 1 + np.random.rand() * 0.2
    sigma = 50 * Ts
    # zero noise
    x = GaussianPulse(time, t0, m0, sigma, noise=0.0)
    assert_almost_equal(np.max(x), m0)
    assert_almost_equal(time[x.argmax()], t0)
    assert np.std(x[: N // 10]) < 1e-10
    # noisy signal
    nstd = 1e-2
    x = GaussianPulse(time, t0, m0, sigma, noise=nstd)
    assert_almost_equal(np.round(np.std(x[: N // 10]) * 100) / 100, nstd)


def test_rect():
    width = N // 4 * Ts
    height = 1 + np.random.rand() * 0.2
    x = rect(time, t0, t0 + width, height, noise=0.0)
    assert_almost_equal(np.max(x), height)
    assert np.max(x[time < t0]) < 1e-10
    assert np.max(x[time > t0 + width]) < 1e-10
    # noisy signal
    nstd = 1e-2
    x = rect(time, t0, t0 + width, height, noise=nstd)
    assert np.round(np.std(x[time < t0]) * 100) / 100 == approx(nstd)


def test_squarepulse():
    height = 1 + np.random.rand() * 0.2
    numpulses = 5
    x = squarepulse(time, height, numpulses, noise=0.0)
    assert_almost_equal(np.max(x), height)


class TestSine:
    # Test the sine signal.
    hi_res_time = get_timestamps(0, 2 * np.pi, 1e-5)

    def test_minimal_call_max_sine(self):
        x = sine(time)
        # Check for minimal callability and that maximum amplitude at
        # timestamps is below default.
        assert np.max(np.abs(x)) <= 1.0

    def test_minimal_call_hi_res_max_sine(self):
        x = sine(self.hi_res_time)
        # Check for minimal callability with high resolution time vector and
        # that maximum amplitude at timestamps is almost equal default.
        assert_almost_equal(np.max(x), 1.0)
        assert_almost_equal(np.min(x), -1.0)

    def test_medium_call_freq_multiples_sine(self):
        # Initialize fixed frequency and number of repetitions.
        np.random.seed(130)
        freq = 100 * np.random.rand()
        rep = 10 * freq
        # Create time vector with timestamps near multiples of frequency.
        fixed_freq_time = get_timestamps(time[0], rep, freq)
        x = sine(fixed_freq_time, freq=freq)
        # Check if signal at multiples of frequency is start value of signal.
        for i_x in x:
            assert_almost_equal(i_x, 0)

    def test_medium_call_max_sine(self):
        # Initialize fixed amplitude.
        np.random.seed(11201)
        amp = np.random.rand()
        x = sine(time, amp=amp)
        # Check for minimal callability and that maximum amplitude at
        # timestamps is below default.
        assert np.max(np.abs(x)) <= amp

    def test_medium_call_hi_res_max_sine(self):
        # Initialize fixed amplitude.
        np.random.seed(1101)
        amp = np.random.rand()
        x = sine(self.hi_res_time, amp=amp)
        # Check for minimal callability with high resolution time vector and
        # that maximum amplitude at timestamps is almost equal default.
        assert_almost_equal(np.max(x), amp)
        assert_almost_equal(np.min(x), -amp)

    def test_full_call_sine(self):
        # Check for all possible calls.
        np.random.seed(1121)
        amp = np.random.rand()
        freq = 100 * np.random.rand()
        noise = np.random.rand()
        sine(time)
        sine(time, amp)
        sine(time, amp=amp)
        sine(time, amp, freq)
        sine(time, amp, freq=freq)
        sine(time, amp=amp, freq=freq)
        sine(time, freq=freq)
        sine(time, amp, freq, noise=noise)
        sine(time, noise=noise)
        sine(time, amp, freq=freq, noise=noise)
        sine(time, freq=freq, noise=noise)
        sine(time, amp=amp, freq=freq, noise=noise)


def test_signal_example(monkeypatch):
    # With this expression we override the matplotlib.pyplot.show method with a
    # lambda expression returning None but only for this one test.
    monkeypatch.setattr(matplotlib.pyplot, "show", lambda: None, raising=True)
    demonstrate_signal()



# -*- coding: utf-8 -*-
"""
The :mod:`PyDynamic.identification` package implements the least-squares fit of an
IIR or FIR digital filter to a given frequency response and the identification of
transfer function models.

# See http://mathmet.org/projects/14SIP08 and
# https://www.github.com/eichstaedtPTB/PyDynamic
"""

from .fit_filter import LSFIR, LSIIR
from .fit_transfer import fit_sos

__all__ = ["LSFIR", "LSIIR", "fit_sos"]



# -*- coding: utf-8 -*-
"""
The propagation of uncertainties via the FIR and IIR formulae alone does not
enable the derivation of credible intervals, because the underlying
distribution remains unknown. The GUM-S2 Monte Carlo method provides a
reference method for the calculation of uncertainties for such cases.

This module implements Monte Carlo methods for the propagation of
uncertainties for digital filtering.

This module contains the following functions:

* :func:`MC`: Standard Monte Carlo method for application of digital filter
* :func:`SMC`: Sequential Monte Carlo method with reduced computer memory requirements
* :func:`UMC`: Update Monte Carlo method for application of digital filters with
  reduced computer memory requirements
* :func:`UMC_generic`: Update Monte Carlo method with reduced computer memory
  requirements
"""

import functools
import math
import multiprocessing
import sys

import numpy as np
import scipy as sp
import scipy.stats as stats
from scipy.interpolate import interp1d
from scipy.signal import lfilter

from ..misc.filterstuff import isstable
from ..misc.tools import progress_bar
from ..misc.noise import ARMA

__all__ = ["MC", "SMC", "UMC", "UMC_generic"]


class Normal_ZeroCorr:
    """Multivariate normal distribution with zero correlation"""

    def __init__(self, loc=np.zeros(1), scale=np.zeros(1)):
        """
        Parameters
        ----------
            loc: np.ndarray, optional
                mean values, default is zero
            scale: np.ndarray, optional
                standard deviations for the elements in loc, default is zero
        """

        if isinstance(loc, np.ndarray) or isinstance(scale, np.ndarray):

            # convert loc to array if necessary
            if not isinstance(loc, np.ndarray):
                self.loc = loc * np.ones(1)
            else:
                self.loc = loc

            # convert scale to arraym if necessary
            if not isinstance(scale, np.ndarray):
                self.scale = scale * np.ones(1)
            else:
                self.scale = scale

            # if one of both (loc/scale) has length one, make it bigger to fit
            # size of the other
            if self.loc.size != self.scale.size:
                Nmax = max(self.loc.size, self.scale.size)

                if self.loc.size == 1 and self.scale.size != 1:
                    self.loc = self.loc * np.ones(Nmax)

                elif self.scale.size == 1 and self.loc.size != 1:
                    self.scale = self.scale * np.ones(Nmax)

                else:
                    raise ValueError(
                        "loc and scale do not have the same dimensions. (And "
                        "none of them has dim == 1)"
                    )

        else:
            raise TypeError(
                "At least one of loc or scale must be of type " "numpy.ndarray."
            )

    def rvs(self, size=1):
        # This function mimics the behavior of the scipy stats package
        return np.tile(self.loc, (size, 1)) + \
               np.random.randn(size, len(self.loc)) * \
               np.tile(self.scale, (size, 1))


def MC(
        x, Ux, b, a, Uab, runs=1000, blow=None, alow=None,
        return_samples=False, shift=0, verbose=True
):
    r"""Standard Monte Carlo method

    Monte Carlo based propagation of uncertainties for a digital filter (b,a)
    with uncertainty matrix :math:`U_{\theta}` for
    :math:`\theta=(a_1,\ldots,a_{N_a},b_0,\ldots,b_{N_b})^T`

    Parameters
    ----------
        x: np.ndarray
            filter input signal
        Ux: float or np.ndarray
            standard deviation of signal noise (float), point-wise standard
            uncertainties or covariance matrix associated with x
        b: np.ndarray
            filter numerator coefficients
        a: np.ndarray
            filter denominator coefficients
        Uab: np.ndarray
            uncertainty matrix :math:`U_\theta`
        runs: int,optional
            number of Monte Carlo runs
        return_samples: bool, optional
            whether samples or mean and std are returned

    If ``return_samples`` is ``False``, the method returns:

    Returns
    -------
        y: np.ndarray
            filter output signal
        Uy: np.ndarray
            uncertainty associated with

    Otherwise the method returns:

    Returns
    -------
        Y: np.ndarray
            array of Monte Carlo results

    References
    ----------
        * Eichstädt, Link, Harris and Elster [Eichst2012]_
    """

    Na = len(a)
    runs = int(runs)

    Y = np.zeros((runs, len(x)))   # set up matrix of MC results
    theta = np.hstack((a[1:], b))  # create the parameter vector from the filter coefficients
    Theta = np.random.multivariate_normal(theta, Uab, runs)  # Theta is small and thus we

    # can draw the full matrix now.
    if isinstance(Ux, np.ndarray):
        if len(Ux.shape) == 1:
            dist = Normal_ZeroCorr(loc=x, scale=Ux)  # non-iid noise w/o correlation
        else:
            dist = stats.multivariate_normal(x, Ux)  # colored noise
    elif isinstance(Ux, float):
        dist = Normal_ZeroCorr(loc=x, scale=Ux)  # iid noise
    else:
        raise NotImplementedError("The supplied type of uncertainty is not implemented")

    unst_count = 0  # Count how often in the MC runs the IIR filter is unstable.
    st_inds = list()
    if verbose:
        sys.stdout.write("MC progress: ")
    for k in range(runs):
        xn = dist.rvs()  # draw filter input signal
        if not blow is None:
            if alow is None:
                alow = 1.0  # FIR low-pass filter
            xn = lfilter(blow, alow, xn)  # low-pass filtered input signal
        bb = Theta[k, Na - 1 :]
        aa = np.hstack((1.0, Theta[k, : Na - 1]))
        if isstable(bb, aa):
            Y[k, :] = lfilter(bb, aa, xn)
            st_inds.append(k)
        else:
            unst_count += 1  # don't apply the IIR filter if it's unstable
        if np.mod(k, 0.1 * runs) == 0 and verbose:
            sys.stdout.write(" %d%%" % (np.round(100.0 * k / runs)))
    if verbose:
        sys.stdout.write(" 100%\n")

    if unst_count > 0:
        print("In %d Monte Carlo %d filters have been unstable" % (runs, unst_count))
        print("These results will not be considered for calculation of mean and " "std")
        print("However, if return_samples is 'True' then ALL samples are " "returned.")

    Y = np.roll(Y, int(shift), axis=1)  # correct for the (known) sample delay

    if return_samples:
        return Y
    else:
        y = np.mean(Y[st_inds, :], axis=0)
        uy = np.cov(Y[st_inds, :], rowvar=False)
        return y, uy


def SMC(
        x, noise_std, b, a, Uab=None, runs=1000, Perc=None, blow=None,
        alow=None, shift=0, return_samples=False, phi=None, theta=None,
        Delta=0.0
):
    r"""Sequential Monte Carlo method

    Sequential Monte Carlo propagation for a digital filter (b,a) with
    uncertainty matrix :math:`U_{\theta}` for
    :math:`\theta=(a_1,\ldots,a_{N_a},b_0,\ldots,b_{N_b})^T`

    Parameters
    ----------
        x: np.ndarray
            filter input signal
        noise_std: float
            standard deviation of signal noise
        b: np.ndarray
            filter numerator coefficients
        a: np.ndarray
            filter denominator coefficients
        Uab: np.ndarray
            uncertainty matrix :math:`U_\theta`
        runs: int, optional
            number of Monte Carlo runs
        Perc: list, optional
            list of percentiles for quantile calculation
        blow: np.ndarray
            optional low-pass filter numerator coefficients
        alow: np.ndarray
            optional low-pass filter denominator coefficients
        shift: int
            integer for time delay of output signals
        return_samples: bool, otpional
            whether to return y and Uy or the matrix Y of MC results
        phi, theta: np.ndarray, optional
            parameters for AR(MA) noise model
            :math:`\epsilon(n)  = \sum_k \phi_k\epsilon(n-k) + \sum_k
            \theta_k w(n-k) + w(n)` with :math:`w(n)\sim N(0,noise_std^2)`
        Delta: float,optional
             upper bound on systematic error of the filter

    If ``return_samples`` is ``False``, the method returns:

    Returns
    -------
        y: np.ndarray
            filter output signal (Monte Carlo mean)
        Uy: np.ndarray
            uncertainties associated with y (Monte Carlo point-wise std)
        Quant: np.ndarray
            quantiles corresponding to percentiles ``Perc`` (if not ``None``)

    Otherwise the method returns:

    Returns
    -------
        Y: np.ndarray
            array of all Monte Carlo results

    References
    ----------
        * Eichstädt, Link, Harris, Elster [Eichst2012]_
    """

    runs = int(runs)

    if isinstance(a, np.ndarray):  # filter order denominator
        Na = len(a) - 1
    else:
        Na = 0
    if isinstance(b, np.ndarray):  # filter order numerator
        Nb = len(b) - 1
    else:
        Nb = 0

    # Initialize noise matrix corresponding to ARMA noise model.
    if isinstance(theta, np.ndarray) or isinstance(theta, float):
        MA = True
        if isinstance(theta, float):
            W = np.zeros((runs, 1))
        else:
            W = np.zeros((runs, len(theta)))
    else:
        MA = False  # no moving average part in noise process

    # Initialize for autoregressive part of noise process.
    if isinstance(phi, np.ndarray) or isinstance(phi, float):
        AR = True
        if isinstance(phi, float):
            E = np.zeros((runs, 1))
        else:
            E = np.zeros((runs, len(phi)))
    else:
        AR = False  # No autoregressive part in noise process.

    # Initialize matrix of low-pass filtered input signal.
    if isinstance(blow, np.ndarray):
        X = np.zeros((runs, len(blow)))
    else:
        X = np.zeros(runs)

    if isinstance(alow, np.ndarray):
        Xl = np.zeros((runs, len(alow) - 1))
    else:
        Xl = np.zeros((runs, 1))

    if Na == 0:  # only FIR filter
        coefs = b
    else:
        coefs = np.hstack((a[1:], b))

    if isinstance(Uab, np.ndarray):  # Monte Carlo draw for filter coefficients
        Coefs = np.random.multivariate_normal(coefs, Uab, runs)
    else:
        Coefs = np.tile(coefs, (runs, 1))

    b0 = Coefs[:, Na]

    if Na > 0:  # filter is IIR
        A = Coefs[:, :Na]
        if Nb > Na:
            A = np.hstack((A, np.zeros((runs, Nb - Na))))
    else:  # filter is FIR -> zero state equations
        A = np.zeros((runs, Nb))

    # Fixed part of state-space model.
    c = Coefs[:, Na + 1 :] - np.multiply(np.tile(b0[:, np.newaxis], (1, Nb)), A)
    States = np.zeros(np.shape(A))  # initialise matrix of states

    calcP = False  # by default no percentiles requested
    if Perc is not None:  # percentiles requested
        calcP = True
        P = np.zeros((len(Perc), len(x)))

    # Initialize outputs.
    y = np.zeros_like(x)
    # Initialize vector of uncorrelated point-wise uncertainties.
    Uy = np.zeros_like(x)

    # Start of the actual MC part.
    print("Sequential Monte Carlo progress", end="")

    for index in np.ndenumerate(x):

        w = np.random.randn(runs) * noise_std  # noise process draw
        if AR and MA:
            E = np.hstack((E.dot(phi) + W.dot(theta) + w, E[:-1]))
            W = np.hstack((w, W[:-1]))
        elif AR:
            E = np.hstack((E.dot(phi) + w, E[:-1]))
        elif MA:
            E = W.dot(theta) + w
            W = np.hstack((w, W[:-1]))
        else:
            w = np.random.randn(runs, 1) * noise_std
            E = w

        if isinstance(alow, np.ndarray):  # apply low-pass filter
            X = np.hstack((x[index] + E, X[:, :-1]))
            Xl = np.hstack(
                (X.dot(blow.T) - Xl[:, : len(alow)].dot(alow[1:]), Xl[:, :-1])
            )
        elif isinstance(blow, np.ndarray):
            X = np.hstack((x[index] + E, X[:, :-1]))
            Xl = X.dot(blow)
        else:
            Xl = x[index] + E

        # Prepare for easier calculations.
        if len(Xl.shape) == 1:
            Xl = Xl[:, np.newaxis]

        # State-space system output.
        Y = (
            np.sum(np.multiply(c, States), axis=1)
            + np.multiply(b0, Xl[:, 0])
            + (np.random.rand(runs) * 2 * Delta - Delta)
        )
        # Calculate state updates.
        Z = -np.sum(np.multiply(A, States), axis=1) + Xl[:, 0]
        # Store state updates and remove old ones.
        States = np.hstack((Z[:, np.newaxis], States[:, :-1]))

        y[index] = np.mean(Y)  # point-wise best estimate
        Uy[index] = np.std(Y)  # point-wise standard uncertainties
        if calcP:
            P[:, index] = sp.stats.mstats.mquantiles(np.asarray(Y), prob=Perc)

        if np.mod(index, np.round(0.1 * len(x))) == 0:
            print(" %d%%" % (np.round(100.0 * index / len(x))), end="")

    print(" 100%")

    # Correct for (known) delay.
    y = np.roll(y, int(shift))
    Uy = np.roll(Uy, int(shift))

    if calcP:
        P = np.roll(P, int(shift), axis=1)
        return y, Uy, P
    else:
        return y, Uy


def UMC(
        x, b, a, Uab, runs=1000, blocksize=8, blow=1.0, alow=1.0, phi=0.0,
        theta=0.0, sigma=1, Delta=0.0, runs_init=100, nbins=1000,
        credible_interval=0.95
):
    """
    Batch Monte Carlo for filtering using update formulae for mean, variance and (approximated) histogram.
    This is a wrapper for the UMC_generic function, specialised on filters

    Parameters
    ----------
        x: np.ndarray, shape (nx, )
            filter input signal
        b: np.ndarray, shape (nbb, )
            filter numerator coefficients
        a: np.ndarray, shape (naa, )
            filter denominator coefficients, normalization (a[0] == 1.0) is assumed
        Uab: np.ndarray, shape (naa + nbb - 1, )
            uncertainty matrix :math:`U_\\theta`
        runs: int, optional
            number of Monte Carlo runs
        blocksize: int, optional
            how many samples should be evaluated for at a time
        blow: float or np.ndarray, optional
            filter coefficients of optional low pass filter
        alow: float or np.ndarray, optional
            filter coefficients of optional low pass filter
        phi: np.ndarray, optional,
            see misc.noise.ARMA noise model
        theta: np.ndarray, optional
            see misc.noise.ARMA noise model
        sigma: float, optional
            see misc.noise.ARMA noise model
        Delta: float, optional
            upper bound of systematic correction due to regularisation (assume uniform distribution)
        runs_init: int, optional
            how many samples to evaluate to form initial guess about limits
        nbins: int, list of int, optional
            number of bins for histogram
        credible_interval: float, optional
            must be in [0,1]
            central credible interval size

    By default, phi, theta, sigma are chosen such, that N(0,1)-noise is added to the input signal.

    Returns
    -------
        y: np.ndarray
            filter output signal
        Uy: np.ndarray
            uncertainty associated with
        y_cred_low: np.ndarray
            lower boundary of credible interval
        y_cred_high: np.ndarray
            upper boundary of credible interval
        happr: dict
            dictionary keys: given nbin
            dictionary values: bin-edges val["bin-edges"], bin-counts val["bin-counts"]

    References
    ----------
        * Eichstädt, Link, Harris, Elster [Eichst2012]_
        * ported to python in 2019-08 from matlab-version of Sascha Eichstaedt (PTB) from 2011-10-12
        * copyright on updating formulae parts is by Peter Harris (NPL)
    """

    # input adjustments and type conversions
    if isinstance(alow, float):
        alow = np.array([alow])
    if isinstance(blow, float):
        blow = np.array([blow])
    if isinstance(nbins, int):
        nbins = [nbins]

    if alow[0] != 1:
        blow = blow / alow[0]
        alow = alow / alow[0]

    # define generic functions to hand over to UMC_generic

    # variate the coefficients of filter as main simulation influence
    ab = np.hstack((a[1:], b))    # create the parameter vector from the filter coefficients (should be named theta, but this name is already used)
    draw_samples = lambda size: np.random.multivariate_normal(ab, Uab, size)

    # how to evaluate functions
    params = {
        "nbb": b.size,
        "x": x,
        "sigma": sigma,
        "blow": blow,
        "alow": alow,
        "Delta": Delta,
        "phi": phi,
        "theta": theta,
    }
    evaluate = functools.partial(_UMCevaluate, **params)

    # run UMC
    y, Uy, happr, _ = UMC_generic(draw_samples, evaluate, runs=runs, blocksize=blocksize, runs_init=runs_init)

    # further post-calculation steps
    y_cred_low = np.zeros((len(nbins), len(y)))
    y_cred_high = np.zeros((len(nbins), len(y)))

    # approximate lower and upper credible quantiles
    for k in range(x.size):
        for m, h in enumerate(happr.values()):
            e = h["bin-edges"][:,k]                 # take all bin-edges
            f = np.append(0, h["bin-counts"][:,k])  # bin count for before first bin is 0
            G = np.cumsum(f)/np.sum(f)

            ## interpolate the cumulated relative bin-count G(e) for the requested credibility interval
            interp_e = interp1d(G, e)

            # save credibility intervals
            y_cred_low[m, k] = interp_e((1 - credible_interval) / 2)
            y_cred_high[m, k] = interp_e((1 + credible_interval) / 2)

    return y, Uy, y_cred_low, y_cred_high, happr


def _UMCevaluate(th, nbb, x, Delta, phi, theta, sigma, blow, alow):
    """
    Calculate system-response of an IIR-filter to some input signal x.

    Parameters
    ----------
    th: numpy.ndarray, shape (naa + nbb -1, )
        coefficients of an IIR filter, :math:`\\theta = [aa[1:], b]`
    nbb: int
        size of bb within th
    x: numpy.ndarray, shape (nx, )
        input signal
    Delta: float
        add noise to output drawn from uniform-distribution U([-Delta, Delta])
    phi: np.ndarray
        see misc.noise.ARMA noise model
    theta: np.ndarray
        see misc.noise.ARMA noise model
    sigma: float
        see misc.noise.ARMA noise model
    blow: float or np.ndarray
        filter coefficients of low pass filter applied to sum of input-signal and ARMA-noise
    alow: float or np.ndarray
        filter coefficients of low pass filter applied to sum of input-signal and ARMA-noise

    ```
    x -----------------+--->[LOWPASS]--->[IIR-FILTER]----+---> y
                       |                                 |
    ARMA(phi,theta) ---'            U([-Delta,Delta]) ---'
    ```
    """

    naa = len(th) - nbb + 1  # theta contains all but the first entry of aa
    aa = np.append(1, th[: naa - 1])  # insert coeff 1 at position 0 to restore aa
    bb = th[naa - 1 :]  # restore bb

    e = ARMA(x.size, phi=phi, theta=theta, std=sigma)

    xlow = lfilter(blow, alow, x + e)
    d = Delta * (2 * np.random.random_sample(size=x.size) - 1 )   # uniform distribution [-Delta, Delta]

    return lfilter(bb, aa, xlow) + d


def UMC_generic(draw_samples, evaluate, runs = 100, blocksize = 8, runs_init = 10, nbins = 100,
                return_samples = False, n_cpu = multiprocessing.cpu_count()):
    """
    Generic Batch Monte Carlo using update formulae for mean, variance and (approximated) histogram.
    Assumes that the input and output of evaluate are numeric vectors (but not necessarily of same dimension).
    If the output of evaluate is multi-dimensional, it will be flattened into 1D. 

    Parameters
    ----------
        draw_samples: function(int nDraws)
            function that draws nDraws from a given distribution / population
            needs to return a list of (multi dimensional) numpy.ndarrays
        evaluate: function(sample)
            function that evaluates a sample and returns the result
            needs to return a (multi dimensional) numpy.ndarray
        runs: int, optional
            number of Monte Carlo runs
        blocksize: int, optional
            how many samples should be evaluated for at a time
        runs_init: int, optional
            how many samples to evaluate to form initial guess about limits
        nbins: int, list of int, optional
            number of bins for histogram
        return_samples: bool, optional
            see return-value of documentation
        n_cpu: int, optional
            number of CPUs to use for multiprocessing, defaults to all available CPUs

    Example
    -------
        draw samples from multivariate normal distribution:
        ``draw_samples = lambda size: np.random.multivariate_normal(x, Ux, size)``

        build a function, that only accepts one argument by masking
        additional kwargs:
        ``evaluate = functools.partial(_UMCevaluate, nbb=b.size, x=x, Delta=Delta, phi=phi, theta=theta, sigma=sigma, blow=blow, alow=alow)``
        ``evaluate = functools.partial(bigFunction, **dict_of_kwargs)``

    By default the method 

    Returns
    -------
        y: np.ndarray
            mean of flattened/raveled simulation output
            i.e.: y = np.ravel(evaluate(sample))
        Uy: np.ndarray
            covariance associated with y
        happr: dict
            dictionary of bin-edges and bin-counts
        output_shape: tuple
            shape of the unraveled simulation output
            can be used to reshape y and np.diag(Uy) into original shape

    If ``return_samples`` is ``True``, the method additionally returns all evaluated samples. 
    This should only be done for testing and debugging reasons, as this removes all memory-improvements of the UMC-method. 

    Returns
    -------
        sims: dict
            dict of samples and corresponding results of every evaluated simulation
            samples and results are saved in their original shape

    References
    ----------
        * Eichstädt, Link, Harris, Elster [Eichst2012]_
    """

    # type-conversions
    if isinstance(nbins, int):
        nbins = [nbins]

    # check if parallel computation is required
    # this allows to circumvent a multiprocessing-problem on windows-machines
    # see: https://github.com/PTB-PSt1/PyDynamic/issues/84
    if n_cpu == 1:
        map_func = map
    else:
        nPool = min(n_cpu, blocksize)
        pool = multiprocessing.Pool(nPool)
        map_func = pool.imap_unordered

    # ------------ preparations for update formulae ------------

    # set up list of MC results
    Y_init = [None] * runs_init

    # init samples to be evaluated
    samples = draw_samples(runs_init)

    # evaluate the initial samples
    for k, result in enumerate(map_func(evaluate, samples)):
        Y_init[k] = result
        progress_bar(k, runs_init, prefix="UMC initialisation:     ")
    print("\n")  # to escape the carriage-return of progress_bar

    # get size of in- and output (was so far not explicitly known)
    input_shape = samples[0].shape
    output_shape = Y_init[0].shape

    # convert to array
    Y_init = np.asarray(Y_init)

    # prepare histograms
    ymin = np.min(Y_init, axis=0).ravel()
    ymax = np.max(Y_init, axis=0).ravel()

    happr = {}
    for nbin in nbins:
        happr[nbin] = {}
        happr[nbin]["bin-edges"] = np.linspace(ymin, ymax, num=nbin+1)           # define bin-edges (generates array for all [ymin,ymax] (assume ymin is already an array))
        happr[nbin]["bin-counts"] = np.zeros((nbin, np.prod(output_shape)))   # init. bin-counts

    # ----------------- run MC block-wise -----------------------

    nblocks = math.ceil(runs / blocksize)

    # remember all evaluated simulations, if wanted
    if return_samples:
        sims = {"samples": np.empty((runs, *input_shape)), "results": np.empty((runs, *output_shape))}

    for m in range(nblocks):
        if m == nblocks:
            curr_block = runs % blocksize
        else:
            curr_block = blocksize

        Y = np.empty((curr_block, np.prod(output_shape)))
        samples = draw_samples(curr_block)

        # evaluate samples in parallel loop
        for k, result in enumerate(map_func(evaluate, samples)):
            Y[k] = result.ravel()

        if m == 0:  # first block
            y = np.mean(Y, axis=0)
            Uy = np.matmul((Y - y).T, (Y - y))

        else:  # updating y and Uy from results of current block
            K0 = m * blocksize
            K_seq = curr_block

            # update mean (formula 7 in [Eichst2012])
            y0 = y
            y = y0 + np.sum(Y - y0, axis=0) / (K0 + K_seq)

            # update covariance (formula 8 in [Eichst2012])
            Uy = ( (K0-1)*Uy + K0*np.outer(y-y0, y-y0) + np.matmul((Y-y).T, (Y-y)) ) / (K0 + K_seq - 1)

        # update histogram values
        for k in range(np.prod(output_shape)):
            for h in happr.values():
                h["bin-counts"][:,k] += np.histogram(Y[:,k], bins = h["bin-edges"][:,k])[0]  # numpy histogram returns (bin-counts, bin-edges)

        ymin = np.min(np.vstack((ymin, Y)), axis=0)
        ymax = np.max(np.vstack((ymax, Y)), axis=0)

        # save results if wanted
        if return_samples:
            block_start = m * blocksize
            block_end = block_start + curr_block
            sims["samples"][block_start:block_end] = samples
            sims["results"][block_start:block_end] = np.asarray([element.reshape(output_shape) for element in Y])

        progress_bar(m*blocksize, runs, prefix="UMC running:            ")  # spaces on purpose, to match length of progress-bar below
    print("\n") # to escape the carriage-return of progress_bar

    # ----------------- post-calculation steps -----------------------

    # replace edge limits by ymin and ymax, resp.
    for h in happr.values():
        h["bin-edges"][0, :] = np.min(np.vstack((ymin, h["bin-edges"][0, :])), axis=0)
        h["bin-edges"][-1, :] = np.min(np.vstack((ymax, h["bin-edges"][-1, :])), axis=0)

    if return_samples:
        return y, Uy, happr, output_shape, sims
    else:
        return y, Uy, happr, output_shape



# -*- coding: utf-8 -*-
""" Perform tests on *misc.filterstuff.isstable*."""

import numpy as np
import pytest

from PyDynamic.misc.filterstuff import isstable


def test_stable():
    # Design a filter, which is known to be stable.
    b = np.array([1, 1])
    a = np.array([1, -0.999999999])
    assert isstable(b, a)


def test_not_stable():
    # Design a filter, which is known to be unstable.
    b = np.array([1, 1])
    a = np.array([-0.999999999, 1])
    assert not isstable(b, a)


def test_error_isstable():
    """Test if isstable raises an error for invalid ftype"""
    b = np.array([1, 1])
    a = np.array([-0.999999999, 1])
    with pytest.raises(ValueError):
        isstable(b, a, ftype="something_invalid")



# -*- coding: utf-8 -*-
"""
.. deprecated:: 1.2.71
    The package *identification* will be combined with the package *deconvolution* and
    renamed to *model_estimation* in the next major release 2.0.0. From version 1.4.1 on
    you should only use the new package :doc:`PyDynamic.model_estimation` instead.

This module contains several functions to carry out a least-squares fit to a
given complex frequency response.

This module for now still contains the following functions:

* :func:`LSIIR`: Least-squares IIR filter fit to a given frequency response
* :func:`LSFIR`: Least-squares fit of a digital FIR filter to a given frequency
  response
"""
import warnings

from ..model_estimation.fit_filter import LSFIR, LSIIR

__all__ = ["LSIIR", "LSFIR"]

warnings.simplefilter("default")
warnings.warn(
    "The module *identification* will be combined with the module "
    "*deconvolution* and renamed to *model_estimation* in the "
    "next major release 2.0.0. From version 1.4.1 on you should only use "
    "the new module *model_estimation* instead.",
    DeprecationWarning,
)



"""
Perform tests on noise methods.
"""

import numpy as np
import scipy
import matplotlib.pyplot as plt
import pytest

# import PyDynamic from local code, not from the (possibly installed) module
import sys
sys.path.append(".")
import PyDynamic.misc.noise as pn

possible_inputs = [2, 0.5, 1, 0, -0.78, -1, -2.0, "blue", "red", "white"]
possible_lengths = [99,100]

def test_get_alpha():
    # typical behavior
    assert pn.get_alpha("red") == -2
    assert pn.get_alpha(2) == 2.0
    assert pn.get_alpha(0.4) == 0.4

    # check error-message for undefined colorname
    with pytest.raises(NotImplementedError) as e:
        pn.get_alpha("nocolor")

def test_white_gaussian():
    w = pn.white_gaussian(10)
    assert len(w) == 10

def test_power_law_noise(visualize=False):

    # check function output for even/uneven N and different alphas
    for color_value in possible_inputs:

        # definitions
        N = np.random.choice(possible_lengths)
        std = 0.5

        # transform w into correlated noise
        w = pn.white_gaussian(N)
        w_color1 = pn.power_law_noise(w = w, std=std, color_value=color_value)
        assert w_color1.shape == (N)
        assert np.all(np.isfinite(w_color1))

        # produce noise of length N
        w_color2 = pn.power_law_noise(N = N, std=std, color_value=color_value)
        assert w_color2.shape == (N)
        assert np.all(np.isfinite(w_color2))

        # visualize the outcome
        if visualize:
            if color_value in [2,1,-2.0]:
                plt.figure("time_alpha=" + str(color_value))
                plt.plot(w)
                plt.plot(w_color1)

                plt.figure("psd_alpha=" + str(color_value))
                plt.psd(w_color1)
                plt.xscale("log")

                #plot.figure("Rxx_matrix_alpha=" + str(color_value))
                #plt.imshow(scipy.linalg.toeplitz(Rxx))

    if visualize:
        plt.show()

def test_power_law_acf():

    # check function output for even/uneven N and different alphas
    for color_value in possible_inputs:

        # definitions
        N = np.random.choice(possible_lengths)
        std = 0.5

        # calculate theoretic covariance
        Rxx = pn.power_law_acf(N, color_value=color_value, std=std)
        assert Rxx.shape == (N,)

        if color_value == 0.0:
            assert Rxx[0] == pytest.approx(std**2)
            assert np.mean(Rxx[1:]) == pytest.approx(0)

def test_ARMA():
    
    # check default parameters (white noise)
    w = pn.ARMA(100)
    assert w.shape == (100,)
    assert np.std(w) == pytest.approx(1.0, abs=0.2)
    assert np.mean(w) == pytest.approx(0.0, abs=0.2)

    # check float, list and numpy.arrays as input values
    phi_list = [2, [2], np.array([1,2])]
    theta_list = [1, [1], np.array([1,2])]

    for theta, phi in zip(theta_list, phi_list):

        # definitions
        N = np.random.choice(possible_lengths)
        std = 0.5

        # calculate ARMA processes
        w = pn.ARMA(100, phi=phi, theta=theta, std=std)
        assert w.shape == (100,)


# -*- coding: utf-8 -*-
"""
    Fit of an IIR filter to a simulated frequency response of a second-order dynamic system.

    Try to play with this script by changing the parameters of the 2nd order system or the order of the fitted filter.

.. seealso:: :mod:`..identification.fit_filter`
"""
import numpy as np
from matplotlib.pyplot import figure, cla, show
from scipy.signal import freqz

import PyDynamic.identification.fit_filter as fit_filter
from PyDynamic.misc.SecondOrderSystem import sos_FreqResp
from PyDynamic.misc.filterstuff import db

# sensor/measuring system
f0 = 36e3           # system resonance frequency in Hz
S0 = 0.124          # system static gain
delta = 0.0055      # system damping

f = np.linspace(0, 80e3, 30)               # frequencies for fitting the system
Hvals = sos_FreqResp(S0, delta, f0, f)      # frequency response of the 2nd order system

# fitting the IIR filter
Fs = 500e3          # sampling frequency
Na = 4; Nb = 4      # IIR filter order (Na - denominator, Nb - numerator)

b, a, tau = fit_filter.LSIIR(Hvals, Na, Nb, f, Fs)      # fit IIR filter to freq response

fplot = np.linspace(0, 80e3, 1000)                # frequency range for the plot
Hc = sos_FreqResp(S0, delta, f0, fplot)           # frequency response of the 2nd order system
Hf = freqz(b, a, 2 * np.pi * fplot / Fs)[1]       # frequency response of the fitted IIR filter
Hf = Hf*np.exp(2j*np.pi*fplot/Fs*tau)             # take into account the filter time delay tau

fig1 = figure(1); cla()
ax1 = fig1.add_subplot(111)
ax1.plot(fplot, db(Hc), "+",fplot, db(Hf))
ax1.legend(('System', 'LSIIR fit'))
ax1.set_xlabel("frequency / Hz",fontsize=18)
ax1.set_ylabel("freq. response amplitude / a.u.",fontsize=18)

fig2 = figure(2); cla()
ax2 = fig2.add_subplot(111)
ax2.plot(fplot, np.angle(Hc), "+",fplot, np.angle(Hf))
ax2.legend(('System', 'LSIIR fit'))
ax2.set_xlabel("frequency / Hz",fontsize=18)
ax2.set_ylabel("freq. response angle / rad",fontsize=18)

show()



# -*- coding: utf-8 -*-
"""
Example for the fit of a second-order transfer function (mass-damper-spring model)
to a set of frequency response values.
"""

from matplotlib.pyplot import *

from PyDynamic.model_estimation.fit_transfer import fit_som
from PyDynamic.misc.SecondOrderSystem import sos_FreqResp
from PyDynamic.misc.tools import make_semiposdef

rst = np.random.RandomState(1)

# sensor/measurement system
S0 = 0.124
uS0 = 0.001
delta = 0.01
udelta = 0.001
f0 = 36
uf0 = 0.5
f = np.linspace(0, f0 * 1.2, 20)

# Monte Carlo for calculation of unc. assoc. with [real(H),imag(H)]
runs = 10000
MCS0 = S0 + rst.randn(runs) * uS0
MCd = delta + rst.randn(runs) * udelta
MCf0 = f0 + rst.randn(runs) * uf0
f = np.linspace(0, 1.2 * f0, 30)

HMC = sos_FreqResp(MCS0, MCd, MCf0, f)

Hc = np.mean(HMC, dtype=complex, axis=1)
H = np.r_[np.real(Hc), np.imag(Hc)]
UH = np.cov(np.r_[np.real(HMC), np.imag(HMC)], rowvar=True)
UH = make_semiposdef(UH)

p, Up, HMC2 = fit_som(f, Hc, UH, scaling=1, MCruns=10000)

figure(1)
errorbar(
    range(1, 4),
    [p[0] * 10, p[1] * 10, p[2] / 10],
    np.sqrt(np.diag(Up)) * np.array([10, 10, 0.1]),
    fmt=".",
)
errorbar(
    np.arange(1, 4) + 0.2,
    [S0 * 10, delta * 10, f0 / 10],
    [uS0 * 10, udelta * 10, uf0 / 10],
    fmt=".",
)
xlim(0, 4)
xticks([1.1, 2.1, 3.1], [r"$S_0$", r"$\delta$", r"$f_0$"])
xlabel("scaled model parameters")

show()



# -*- coding: utf-8 -*-
"""
The module :mod:`PyDynamic.model_estimation.fit_filter` contains several functions to
carry out a least-squares fit to a given complex frequency response and the design of
digital deconvolution filters by least-squares fitting to the reciprocal of a given
frequency response each with associated uncertainties.

This module contains the following functions:

* :func:`LSIIR`: Least-squares IIR filter fit to a given frequency response
* :func:`LSFIR`: Least-squares fit of a digital FIR filter to a given frequency
  response
* :func:`invLSFIR`: Least-squares fit of a digital FIR filter to the reciprocal of a
  given frequency response.
* :func:`invLSFIR_unc`: Design of FIR filter as fit to reciprocal of frequency response
  values with uncertainty
* :func:`invLSFIR_uncMC`: Design of FIR filter as fit to reciprocal of frequency
  response values with uncertainty via Monte Carlo
* :func:`invLSIIR`: Design of a stable IIR filter as fit to reciprocal of frequency
  response values
* :func:`invLSIIR_unc`: Design of a stable IIR filter as fit to reciprocal of frequency
  response values with uncertainty

"""
import warnings

import numpy as np
import scipy.signal as dsp

from ..misc.filterstuff import grpdelay, mapinside

__all__ = [
    "LSIIR",
    "LSFIR",
    "invLSFIR",
    "invLSFIR_unc",
    "invLSIIR",
    "invLSIIR_unc",
    "invLSFIR_uncMC"
    ]


def _fitIIR(
    Hvals: np.ndarray,
    tau: int,
    w: np.ndarray,
    E: np.ndarray,
    Na: int,
    Nb: int,
    inv: bool = False
    ):
    """The actual fitting routing for the least-squares IIR filter.

    Parameters
    ----------
        Hvals :  (M,) np.ndarray
            (complex) frequency response values
        tau : integer
            initial estimate of time delay
        w : np.ndarray
            :math:`2 * np.pi * f / Fs`
        E : np.ndarray
            :math:`np.exp(-1j * np.dot(w[:, np.newaxis], Ns.T))`
        Nb : int
            numerator polynomial order
        Na : int
            denominator polynomial order
        inv : bool, optional
            If True the least-squares fitting is performed for the reciprocal, if False
            for the actual frequency response

    Returns
    -------
        b, a : IIR filter coefficients as numpy arrays
    """
    exponent = -1 if inv else 1
    Ea = E[:, 1:Na + 1]
    Eb = E[:, :Nb + 1]
    Htau = np.exp(-1j * w * tau) * Hvals ** exponent
    HEa = np.dot(np.diag(Htau), Ea)
    D = np.hstack((HEa, -Eb))
    Tmp1 = np.real(np.dot(np.conj(D.T), D))
    Tmp2 = np.real(np.dot(np.conj(D.T), -Htau))
    ab = np.linalg.lstsq(Tmp1, Tmp2, rcond=None)[0]
    a = np.hstack((1.0, ab[:Na]))
    b = ab[Na:]
    return b, a


def LSIIR(Hvals, Nb, Na, f, Fs, tau=0, justFit=False):
    """Least-squares IIR filter fit to a given frequency response.
    
    This method uses Gauss-Newton non-linear optimization and pole
    mapping for filter stabilization
    
    Parameters
    ----------
        Hvals:   numpy array of (complex) frequency response values of shape (M,)
        Nb:      integer numerator polynomial order
        Na:      integer denominator polynomial order
        f:       numpy array of frequencies at which Hvals is given of shape
        (M,)
        Fs:      sampling frequency
        tau:     integer initial estimate of time delay
        justFit: boolean, when true then no stabilization is carried out
    
    Returns
    -------
        b,a:    IIR filter coefficients as numpy arrays
        tau:    filter time delay in samples

    References
    ----------
    * Eichstädt et al. 2010 [Eichst2010]_
    * Vuerinckx et al. 1996 [Vuer1996]_

    """

    print("\nLeast-squares fit of an order %d digital IIR filter" % max(Nb, Na))
    print("to a frequency response given by %d values.\n" % len(Hvals))

    w = 2 * np.pi * f / Fs
    Ns = np.arange(0, max(Nb, Na) + 1)[:, np.newaxis]
    E = np.exp(-1j * np.dot(w[:, np.newaxis], Ns.T))

    b, a = _fitIIR(Hvals, tau, w, E, Na, Nb, inv=False)

    if justFit:
        print("Calculation done. No stabilization requested.")
        if np.count_nonzero(np.abs(np.roots(a)) > 1) > 0:
            print("Obtained filter is NOT stable.")
        sos = np.sum(
            np.abs((dsp.freqz(b, a, 2 * np.pi * f / Fs)[1] - Hvals) ** 2))
        print("Final sum of squares = %e" % sos)
        tau = 0
        return b, a, tau

    if np.count_nonzero(np.abs(np.roots(a)) > 1) > 0:
        stable = False
    else:
        stable = True

    maxiter = 50

    astab = mapinside(a)
    run = 1

    while stable is not True and run < maxiter:
        g1 = grpdelay(b, a, Fs)[0]
        g2 = grpdelay(b, astab, Fs)[0]
        tau = np.ceil(tau + np.median(g2 - g1))

        b, a = _fitIIR(Hvals, tau, w, E, Na, Nb, inv=False)
        if np.count_nonzero(np.abs(np.roots(a)) > 1) > 0:
            astab = mapinside(a)
        else:
            stable = True
        run = run + 1

    if np.count_nonzero(np.abs(np.roots(a)) > 1) > 0:
        print("Caution: The algorithm did NOT result in a stable IIR filter!")
        print(
            "Maybe try again with a higher value of tau0 or a higher filter "
            "order?")

    print(
        "Least squares fit finished after %d iterations (tau=%d)." % (run, tau))
    Hd = dsp.freqz(b, a, 2 * np.pi * f / Fs)[1]
    Hd = Hd * np.exp(1j * 2 * np.pi * f / Fs * tau)
    res = np.hstack(
        (np.real(Hd) - np.real(Hvals), np.imag(Hd) - np.imag(Hvals)))
    rms = np.sqrt(np.sum(res ** 2) / len(f))
    print("Final rms error = %e \n\n" % rms)

    return b, a, int(tau)


def LSFIR(H, N, tau, f, Fs, Wt=None):
    """
    Least-squares fit of a digital FIR filter to a given frequency response.

    Parameters
    ----------
        H : (complex) frequency response values of shape (M,)
        N : FIR filter order
        tau : delay of filter
        f : frequencies of shape (M,)
        Fs : sampling frequency of digital filter
        Wt : (optional) vector of weights of shape (M,) or shape (M,M)

    Returns
    -------
        filter coefficients bFIR (ndarray) of shape (N+1,)

    """

    print("\nLeast-squares fit of an order %d digital FIR filter to the" % N)
    print("reciprocal of a frequency response given by %d values.\n" % len(H))

    H = H[:, np.newaxis]

    w = 2 * np.pi * f / Fs
    w = w[:, np.newaxis]

    ords = np.arange(N + 1)[:, np.newaxis]
    ords = ords.T

    E = np.exp(-1j * np.dot(w, ords))

    if Wt is not None:
        if len(np.shape(Wt)) == 2:  # is matrix
            weights = np.diag(Wt)
        else:
            weights = np.eye(len(f)) * Wt
        X = np.vstack(
            [np.real(np.dot(weights, E)), np.imag(np.dot(weights, E))])
    else:
        X = np.vstack([np.real(E), np.imag(E)])

    H = H * np.exp(1j * w * tau)
    iRI = np.vstack([np.real(1.0 / H), np.imag(1.0 / H)])

    bFIR, res = np.linalg.lstsq(X, iRI)[:2]

    if not isinstance(res, np.ndarray):
        print(
            "Calculation of FIR filter coefficients finished with residual "
            "norm %e" % res)

    return np.reshape(bFIR, (N + 1,))

def invLSFIR(H, N, tau, f, Fs, Wt=None):
    """	Least-squares fit of a digital FIR filter to the reciprocal of a given
    frequency response.

    Parameters
    ----------
        H: np.ndarray of shape (M,) and dtype complex
            frequency response values
        N: int
            FIR filter order
        tau: float
            delay of filter
        f: np.ndarray of shape (M,)
            frequencies
        Fs: float
            sampling frequency of digital filter
        Wt: np.ndarray of shape (M,) - optional
            vector of weights

    Returns
    -------
        bFIR: np.ndarray of shape (N,)
            filter coefficients

    References
    ----------
        * Elster and Link [Elster2008]_

    .. see_also ::mod::`PyDynamic.uncertainty.propagate_filter.FIRuncFilter`

    """

    print("\nLeast-squares fit of an order %d digital FIR filter to the" % N)
    print("reciprocal of a frequency response given by %d values.\n" % len(H))

    H = H[:, np.newaxis]  # extend to matrix-like for simplified algebra

    w = 2 * np.pi * f / Fs  # set up radial frequencies
    w = w[:, np.newaxis]

    ords = np.arange(N + 1)[:, np.newaxis]  # set up design matrix
    ords = ords.T

    E = np.exp(-1j * np.dot(w, ords))

    if Wt is not None:  # set up weighted design matrix if necessary
        if len(np.shape(Wt)) == 2:  # is matrix
            weights = np.diag(Wt)
        else:
            weights = np.eye(len(f)) * Wt
        X = np.vstack([np.real(np.dot(weights, E)), np.imag(np.dot(weights, E))])
    else:
        X = np.vstack([np.real(E), np.imag(E)])

    Hs = H * np.exp(1j * w * tau)  # apply time delay for improved fit quality
    iRI = np.vstack([np.real(1.0 / Hs), np.imag(1.0 / Hs)])

    bFIR, res = np.linalg.lstsq(X, iRI)[:2]  # the actual fitting

    if (not isinstance(res, np.ndarray)) or (len(res) == 1):  # summarise results
        print(
            "Calculation of FIR filter coefficients finished with residual "
            "norm %e" % res
        )
        Hd = dsp.freqz(bFIR, 1, 2 * np.pi * f / Fs)[1]
        Hd = Hd * np.exp(1j * 2 * np.pi * f / Fs * tau)
        res = np.hstack((np.real(Hd) - np.real(H), np.imag(Hd) - np.imag(H)))
        rms = np.sqrt(np.sum(res ** 2) / len(f))
        print("Final rms error = %e \n\n" % rms)

    return bFIR.flatten()


def invLSFIR_unc(H, UH, N, tau, f, Fs, wt=None, verbose=True, trunc_svd_tol=None):
    """Design of FIR filter as fit to reciprocal of frequency response values
    with uncertainty


    Least-squares fit of a digital FIR filter to the reciprocal of a
    frequency response
    for which associated uncertainties are given for its real and imaginary
    part.
    Uncertainties are propagated using a truncated svd and linear matrix
    propagation.

    Parameters
    ----------
        H: np.ndarray of shape (M,)
            frequency response values
        UH: np.ndarray of shape (2M,2M)
            uncertainties associated with the real and imaginary part
        N: int
            FIR filter order
        tau: float
            delay of filter
        f: np.ndarray of shape (M,)
            frequencies
        Fs: float
            sampling frequency of digital filter
        wt: np.ndarray of shape (2M,) - optional
            array of weights for a weighted least-squares method
        verbose: bool, optional
            whether to print statements to the command line
        trunc_svd_tol: float
            lower bound for singular values to be considered for pseudo-inverse

    Returns
    -------
        b: np.ndarray of shape (N+1,)
            filter coefficients of shape
        Ub: np.ndarray of shape (N+1,N+1)
            uncertainties associated with b

    References
    ----------
        * Elster and Link [Elster2008]_
    """

    if verbose:
        print("\nLeast-squares fit of an order %d digital FIR filter to the" % N)
        print("reciprocal of a frequency response given by %d values" % len(H))
        print("and propagation of associated uncertainties.")

    # Step 1: Propagation of uncertainties to reciprocal of frequency response
    runs = 10000
    Nf = len(f)

    if not len(H) == UH.shape[0]:
        # Assume that H is given as complex valued frequency response.
        RI = np.hstack((np.real(H), np.imag(H)))
    else:
        RI = H.copy()
        H = H[:Nf] + 1j * H[Nf:]
    HRI = np.random.multivariate_normal(RI, UH, runs)  # random draws of real,imag of
    # freq response values
    omtau = 2 * np.pi * f / Fs * tau

    # Vectorized Monte Carlo for propagation to inverse
    absHMC = HRI[:, :Nf] ** 2 + HRI[:, Nf:] ** 2
    HiMC = np.hstack(((HRI[:, :Nf] * np.tile(np.cos(omtau), (runs, 1)) +
                       HRI[:, Nf:] * np.tile(np.sin(omtau), (runs, 1))) /
                      absHMC,
                      (HRI[:, Nf:] * np.tile(np.cos(omtau), (runs, 1)) -
                       HRI[:, :Nf] * np.tile(np.sin(omtau), (runs, 1))) /
                      absHMC))
    UiH = np.cov(HiMC, rowvar=False)

    # Step 2: Fit filter coefficients and evaluate uncertainties
    if isinstance(wt, np.ndarray):
        if wt.shape != np.diag(UiH).shape[0]:
            raise ValueError("User-defined weighting has wrong dimension.")
    else:
        wt = np.ones(2 * Nf)

    E = np.exp(-1j * 2 * np.pi * np.dot(f[:, np.newaxis] / Fs,
                                        np.arange(N + 1)[:, np.newaxis].T))
    X = np.vstack((np.real(E), np.imag(E)))
    X = np.dot(np.diag(wt), X)
    Hm = H * np.exp(1j * 2 * np.pi * f / Fs * tau)
    Hri = np.hstack((np.real(1.0 / Hm), np.imag(1.0 / Hm)))

    u, s, v = np.linalg.svd(X, full_matrices=False)
    if isinstance(trunc_svd_tol, float):
        s[s < trunc_svd_tol] = 0.0
    StSInv = np.zeros_like(s)
    StSInv[s > 0] = s[s > 0] ** (-2)

    M = np.dot(np.dot(np.dot(v.T, np.diag(StSInv)), np.diag(s)), u.T)

    bFIR = np.dot(M, Hri[:, np.newaxis])  # actual fitting
    UbFIR = np.dot(np.dot(M, UiH), M.T)  # evaluation of uncertainties

    bFIR = bFIR.flatten()

    if verbose:
        Hd = dsp.freqz(bFIR, 1, 2 * np.pi * f / Fs)[1]
        Hd = Hd * np.exp(1j * 2 * np.pi * f / Fs * tau)
        res = np.hstack((np.real(Hd) - np.real(H), np.imag(Hd) - np.imag(H)))
        rms = np.sqrt(np.sum(res ** 2) / len(f))
        print("Final rms error = %e \n\n" % rms)

    return bFIR, UbFIR


def invLSFIR_uncMC(H, UH, N, tau, f, Fs, verbose=True):
    """Design of FIR filter as fit to reciprocal of frequency response values
    with uncertainty

    Least-squares fit of a FIR filter to the reciprocal of a frequency response
    for which associated uncertainties are given for its real and imaginary
    parts.
    Uncertainties are propagated using a Monte Carlo method. This method may
    help in cases where
    the weighting matrix or the Jacobian are ill-conditioned, resulting in
    false uncertainties
    associated with the filter coefficients.

    Parameters
    ----------
        H: np.ndarray of shape (M,) and dtype complex
            frequency response values
        UH: np.ndarray of shape (2M,2M)
            uncertainties associated with the real and imaginary part of H
        N: int
            FIR filter order
        tau: int
            time delay of filter in samples
        f: np.ndarray of shape (M,)
            frequencies corresponding to H
        Fs: float
            sampling frequency of digital filter
        verbose: bool, optional
            whether to print statements to the command line

    Returns
    -------
        b: np.ndarray of shape (N+1,)
            filter coefficients of shape
        Ub: np.ndarray of shape (N+1, N+1)
            uncertainties associated with b

    References
    ----------
        * Elster and Link [Elster2008]_
    """

    if verbose:
        print("\nLeast-squares fit of an order %d digital FIR filter to the" % N)
        print("reciprocal of a frequency response given by %d values" % len(H))
        print("and propagation of associated uncertainties.")

    # Step 1: Propagation of uncertainties to reciprocal of frequency response
    runs = 10000
    HRI = np.random.multivariate_normal(np.hstack((np.real(H), np.imag(H))), UH, runs)

    # Step 2: Fitting the filter coefficients
    E = np.exp(-1j * 2 * np.pi * np.dot(f[:, np.newaxis] / Fs,
                                        np.arange(N + 1)[:, np.newaxis].T))
    X = np.vstack((np.real(E), np.imag(E)))

    Nf = len(f)
    bF = np.zeros((N + 1, runs))
    resn = np.zeros((runs,))
    for k in range(runs):
        Hk = HRI[k, :Nf] + 1j * HRI[k, Nf:]
        Hkt = Hk * np.exp(1j * 2 * np.pi * f / Fs * tau)
        iRI = np.hstack([np.real(1.0 / Hkt), np.imag(1.0 / Hkt)])
        bF[:, k], res = np.linalg.lstsq(X, iRI)[:2]
        resn[k] = np.linalg.norm(res)

    bFIR = np.mean(bF, axis=1)
    UbFIR = np.cov(bF, rowvar=True)

    return bFIR, UbFIR


def invLSIIR(Hvals, Nb, Na, f, Fs, tau, justFit=False, verbose=True):
    """Design of a stable IIR filter as fit to reciprocal of frequency
    response values

    Least-squares fit of a digital IIR filter to the reciprocal of a given set
    of frequency response values using the equation-error method and
    stabilization
    by pole mapping and introduction of a time delay.

    Parameters
    ----------
        Hvals: np.ndarray of shape (M,) and dtype complex
            frequency response values.
        Nb: int
            order of IIR numerator polynomial.
        Na: int
            order of IIR denominator polynomial.
        f: np.ndarray of shape (M,)
            frequencies corresponding to Hvals
        Fs: float
            sampling frequency for digital IIR filter.
        tau: float
            initial estimate of time delay for filter stabilization.
        justFit: bool
            if True then no stabilization is carried out.
        verbose: bool
            If True print some more detail about input parameters.

    Returns
    -------
        b : np.ndarray
            The IIR filter numerator coefficient vector in a 1-D sequence
        a : np.ndarray
            The IIR filter denominator coefficient vector in a 1-D sequence
        tau : int
            time delay (in samples)

    References
    ----------
        * Eichstädt, Elster, Esward, Hessling [Eichst2010]_

    """
    from numpy import count_nonzero, roots, ceil, median

    if verbose:
        print(
            "\nLeast-squares fit of an order %d digital IIR filter to the" % max(Nb, Na)
        )
        print("reciprocal of a frequency response given by %d values.\n" % len(Hvals))

    w = 2 * np.pi * f / Fs
    Ns = np.arange(0, max(Nb, Na) + 1)[:, np.newaxis]
    E = np.exp(-1j * np.dot(w[:, np.newaxis], Ns.T))

    bi, ai = _fitIIR(Hvals, tau, w, E, Na, Nb, inv=True)

    if justFit:  # no uncertainty evaluation
        return bi, ai

    if count_nonzero(abs(roots(ai)) > 1) > 0:
        stable = False
    else:
        stable = True

    maxiter = 50

    astab = mapinside(ai)  # stabilise filter
    run = 1

    while stable is not True and run < maxiter:  # shift delay such that filter
        # becomes stable
        g1 = grpdelay(bi, ai, Fs)[0]
        g2 = grpdelay(bi, astab, Fs)[0]
        tau = ceil(tau + median(g2 - g1))

        bi, ai = _fitIIR(Hvals, tau, w, E, Na, Nb, inv=True)
        if count_nonzero(abs(roots(ai)) > 1) > 0:
            astab = mapinside(ai)
        else:
            stable = True
        run = run + 1

    if count_nonzero(abs(roots(ai)) > 1) > 0 and verbose:
        print("Caution: The algorithm did NOT result in a stable IIR filter!")
        print(
            "Maybe try again with a higher value of tau0 or a higher filter order?"
        )

    if verbose:
        print("Least squares fit finished after %d iterations (tau=%d).\n" % (run, tau))
        Hd = dsp.freqz(bi, ai, 2 * np.pi * f / Fs)[1]
        Hd = Hd * np.exp(1j * 2 * np.pi * f / Fs * tau)
        res = np.hstack((np.real(Hd) - np.real(Hvals), np.imag(Hd) - np.imag(Hvals)))
        rms = np.sqrt(np.sum(res ** 2) / len(f))
        print("Final rms error = %e \n\n" % rms)

    return bi, ai, int(tau)


def invLSIIR_unc(H, UH, Nb, Na, f, Fs, tau=0):
    """Design of stabel IIR filter as fit to reciprocal of given frequency
    response with uncertainty

    Least-squares fit of a digital IIR filter to the reciprocal of a given set
    of frequency response values with given associated uncertainty.
    Propagation of uncertainties is
    carried out using the Monte Carlo method.

    Parameters
    ----------

        H: np.ndarray of shape (M,) and dtype complex
            frequency response values.
        UH: np.ndarray of shape (2M,2M)
            uncertainties associated with real and imaginary part of H
        Nb: int
            order of IIR numerator polynomial.
        Na: int
            order of IIR denominator polynomial.
        f: np.ndarray of shape (M,)
            frequencies corresponding to H
        Fs: float
            sampling frequency for digital IIR filter.
        tau: float
            initial estimate of time delay for filter stabilization.

    Returns
    -------
        b,a: np.ndarray
            IIR filter coefficients
        tau: int
            time delay (in samples)
        Uba: np.ndarray of shape (Nb+Na+1, Nb+Na+1)
            uncertainties associated with [a[1:],b]

    References
    ----------
        * Eichstädt, Elster, Esward and Hessling [Eichst2010]_

    .. seealso:: :mod:`PyDynamic.uncertainty.propagate_filter.IIRuncFilter`
                 :mod:`PyDynamic.model_estimation.fit_filter.invLSIIR`
    """

    runs = 1000

    print("\nLeast-squares fit of an order %d digital IIR filter to the" % max(Nb, Na))
    print("reciprocal of a frequency response given by %d values.\n" % len(H))
    print(
        "Uncertainties of the filter coefficients are evaluated using\n"
        "the GUM S2 Monte Carlo method with %d runs.\n" % runs
    )

    # Step 1: Propagation of uncertainties to frequency response
    HRI = np.random.multivariate_normal(np.hstack((np.real(H), np.imag(H))), UH, runs)
    HH = HRI[:, : len(f)] + 1j * HRI[:, len(f) :]

    # Step 2: Fit filter and evaluate uncertainties (Monte Carlo method)
    AB = np.zeros((runs, Nb + Na + 1))
    Tau = np.zeros((runs,))
    for k in range(runs):
        bi, ai, Tau[k] = invLSIIR(HH[k, :], Nb, Na, f, Fs, tau, verbose=False)
        AB[k, :] = np.hstack((ai[1:], bi))

    bi = np.mean(AB[:, Na:], axis=0)
    ai = np.hstack((np.array([1.0]), np.mean(AB[:, :Na], axis=0)))
    Uab = np.cov(AB, rowvar=False)
    tau = np.mean(Tau)
    return bi, ai, tau, Uab



# -*- coding: utf-8 -*-
"""
.. deprecated:: 2.0.0
    The module :mod:`PyDynamic.uncertainty.interpolation` will be renamed to
    :mod:`PyDynamic.uncertainty.interpolate` in the next major release 2.0.0. From
    version 1.4.3 on you should only use the new module instead.

The :mod:`PyDynamic.uncertainty.interpolation` module implements methods for
the propagation of uncertainties in the application of standard interpolation methods
as provided by :class:`scipy.interpolate.interp1d`.

This module for now still contains the following function:

* :func:`interp1d_unc`: Interpolate arbitrary time series considering the associated
  uncertainties
"""
import warnings

from .interpolate import interp1d_unc

__all__ = ["interp1d_unc"]

warnings.simplefilter("default")
warnings.warn(
    "The module :mod:`PyDynamic.uncertainty.interpolation` will be renamed to "
    ":mod:`PyDynamic.uncertainty.interpolate` in the next major release 2.0.0. From "
    "version 1.4.3 on you should only use the new module instead.",
    PendingDeprecationWarning,
)



from typing import Dict, Optional, Tuple, Union

import hypothesis.extra.numpy as hnp
import hypothesis.strategies as st
import numpy as np
from hypothesis import assume, given
from hypothesis.strategies import composite
from numpy.testing import assert_allclose
from pytest import raises

from PyDynamic.uncertainty.interpolation import interp1d_unc


@composite
def timestamps_values_uncertainties_kind(
    draw,
    min_count: Optional[int] = 2,
    max_count: Optional[int] = None,
    kind_tuple: Optional[Tuple[str]] = ("linear", "previous", "next", "nearest"),
    sorted_timestamps: Optional[bool] = True,
    extrapolate: Optional[Union[bool, str]] = False,
    restrict_fill_value: Optional[str] = None,
    restrict_fill_unc: Optional[str] = None,
    returnC: Optional[bool] = False,
) -> Dict[str, Union[np.ndarray, str]]:
    """Set custom strategy for _hypothesis_ to draw desired input from

    Parameters
    ----------
        draw : callable
            this is a hypothesis internal callable to actually draw from provided
            strategies
        min_count : int
            the minimum number of elements expected inside the arrays of timestamps
             (or frequencies), measurement values and associated uncertainties
        max_count : int
            the maximum number of elements expected inside the arrays of timestamps
            (or frequencies), measurement values and associated uncertainties
        kind_tuple : tuple(str), optional
            the tuple of strings out of "linear", "previous", "next", "nearest",
            "spline", "least-squares" from which the strategy for the
            kind randomly chooses. Defaults to the valid options "linear",
            "previous", "next", "nearest"
        sorted_timestamps : bool
            if True the timestamps (or frequencies) are guaranteed to be in ascending
            order, if False they still might be by coincidence or not
        extrapolate : bool or str, optional
            If True the interpolation timestamps (or frequencies) are generated such
            that extrapolation is necessary by guarantying at least one of the
            interpolation timestamps (or frequencies) outside the original bounds
            and accordingly setting appropriate values for `fill_value` and
            `bounds_error = False`. If False each element of t_new is guaranteed to
            lie within the range of t. Can be set to "above" or "below" to guarantee
            at least one element of t_new to lie either below or above the bounds of t.
        restrict_fill_value : str, optional
            String specifying the desired strategy for drawing a fill_value. One of
            "float", "tuple", "str", "nan" to guarantee either a float, a tuple of
            two floats, the string "extrapolate" or np.nan.
        restrict_fill_unc : str, optional
            Same as fill_value, but just for the uncertainties.
        returnC : bool, optional
            If True we request the sensitivities to be returned. If False we do not
            request them. Defaults to False.

    Returns
    -------
        A dict containing the randomly generated expected input parameters for
        `interp1d_unc()`
    """

    def draw_fill_values(strategy_spec: str):
        """Little helper to find proper strategy for efficient testing.

        Parameters
        ----------
            strategy_spec : str
                String specifying the desired strategy for drawing a fill_value. One of
                "float", "tuple", "str", "nan" to guarantee either a float, a tuple of
                two floats, the string "extrapolate" or np.nan.

        Returns
        -------
            The drawn sample to match desired fill_value.
        """
        float_strategy = st.floats(**float_generic_params)
        tuple_strategy = st.tuples(float_strategy, float_strategy)
        string_strategy = st.just("extrapolate")
        nan_strategy = st.just(np.nan)
        if strategy_spec == "float":
            fill_strategy = float_strategy
        elif strategy_spec == "tuple":
            fill_strategy = tuple_strategy
        elif strategy_spec == "str":
            fill_strategy = string_strategy
        elif strategy_spec == "nan":
            fill_strategy = nan_strategy
        else:
            fill_strategy = st.one_of(
                float_strategy, tuple_strategy, string_strategy, nan_strategy
            )
        return draw(fill_strategy)

    # Set the maximum absolute value for floats to be really unique in calculations.
    float_abs_max = 1e64
    # Set generic float parameters.
    float_generic_params = {
        "allow_nan": False,
        "allow_infinity": False,
    }
    # Set all common parameters for timestamps (or frequencies), measurements values
    # and associated uncertainties.
    shape_for_timestamps = hnp.array_shapes(
        max_dims=1, min_side=min_count, max_side=max_count
    )
    strategy_params = {
        "dtype": np.float,
        "shape": shape_for_timestamps,
        "elements": st.floats(
            min_value=-float_abs_max, max_value=float_abs_max, **float_generic_params
        ),
        "unique": True,
    }
    # Draw "original" timestamps (or frequencies).
    t = draw(hnp.arrays(**strategy_params))
    # Sort timestamps (or frequencies) in ascending order.
    if sorted_timestamps:
        ind = np.argsort(t)
        t = t[ind]

    # Reuse "original" timestamps (or frequencies) shape for measurements values and
    # associated uncertainties and draw both.
    strategy_params["shape"] = np.shape(t)
    y = draw(hnp.arrays(**strategy_params))
    uy = draw(hnp.arrays(**strategy_params))

    # Reset shape for interpolation timestamps (or frequencies).
    strategy_params["shape"] = shape_for_timestamps
    # Look up minimum and maximum of original timestamps (or frequencies) just once.
    t_min = np.min(t)
    t_max = np.max(t)

    if not extrapolate:
        # In case we do not want to extrapolate, use range of "original" timestamps (or
        # frequencies) as boundaries.
        strategy_params["elements"] = st.floats(
            min_value=t_min, max_value=t_max, **float_generic_params
        )
        fill_value = fill_unc = np.nan
        bounds_error = True
    else:
        # In case we want to extrapolate, draw some fill values for the
        # out-of-bounds range. Those will be either single floats or a 2-tuple of
        # floats or the special value "extrapolate".
        fill_value = draw_fill_values(restrict_fill_value)
        fill_unc = draw_fill_values(restrict_fill_unc)
        bounds_error = False

    # Draw interpolation timestamps (or frequencies).
    t_new = draw(hnp.arrays(**strategy_params))

    if extrapolate:
        # In case we want to extrapolate, make sure we actually do after having drawn
        # the timestamps (or frequencies) not to randomly have drawn values inside
        # original bounds and if even more constraints are given ensure those.
        assume(np.min(t_new) < np.min(t) or np.max(t_new) > np.max(t))
        if extrapolate == "above":
            assume(np.max(t_new) > np.max(t))
        else:
            assume(np.min(t_new) < np.min(t))

    kind = draw(st.sampled_from(kind_tuple))
    assume_sorted = sorted_timestamps
    return {
        "t_new": t_new,
        "t": t,
        "y": y,
        "uy": uy,
        "kind": kind,
        "fill_value": fill_value,
        "fill_unc": fill_unc,
        "bounds_error": bounds_error,
        "assume_sorted": assume_sorted,
        "returnC": returnC,
    }


@given(timestamps_values_uncertainties_kind())
def test_usual_call(interp_inputs):
    t_new, y_new, uy_new = interp1d_unc(**interp_inputs)
    # Check the equal dimensions of the minimum calls output.
    assert len(t_new) == len(y_new) == len(uy_new)


@given(timestamps_values_uncertainties_kind())
def test_wrong_input_length_y_call_interp1d_unc(interp_inputs):
    # Check erroneous calls with unequally long inputs.
    interp_inputs["y"] = np.tile(interp_inputs["y"], 2)
    with raises(ValueError):
        interp1d_unc(**interp_inputs)


@given(timestamps_values_uncertainties_kind())
def test_wrong_input_length_uy_call_interp1d_unc(interp_inputs):
    # Check erroneous calls with unequally long inputs.
    interp_inputs["uy"] = np.tile(interp_inputs["uy"], 2)
    with raises(ValueError):
        interp1d_unc(**interp_inputs)


@given(timestamps_values_uncertainties_kind(kind_tuple=("previous", "next", "nearest")))
def test_trivial_in_interp1d_unc(interp_inputs):
    y_new, uy_new = interp1d_unc(**interp_inputs)[1:3]
    # Check if all 'interpolated' values are present in the actual values.
    assert np.all(np.isin(y_new, interp_inputs["y"]))
    assert np.all(np.isin(uy_new, interp_inputs["uy"]))


@given(timestamps_values_uncertainties_kind(kind_tuple=["linear"]))
def test_linear_in_interp1d_unc(interp_inputs):
    y_new, uy_new = interp1d_unc(**interp_inputs)[1:3]
    # Check if all interpolated values lie in the range of the original values.
    assert np.all(np.min(interp_inputs["y"]) <= y_new)
    assert np.all(np.max(interp_inputs["y"]) >= y_new)


@given(timestamps_values_uncertainties_kind(extrapolate=True))
def test_extrapolate_interp1d_unc(interp_inputs):
    # Check that extrapolation is executable in general.
    assert interp1d_unc(**interp_inputs)


@given(
    timestamps_values_uncertainties_kind(
        sorted_timestamps=True, extrapolate="below", restrict_fill_value="str"
    )
)
def test_extrapolate_below_without_fill_value_interp1d_unc(interp_inputs):
    # Deal with those cases where at least one of t_new is below the minimum of t and
    # fill_value=="extrapolate", which means constant extrapolation from the boundaries.
    y_new = interp1d_unc(**interp_inputs)[1]
    # Check that extrapolation works, meaning in the present case, that the boundary
    # value of y is taken for all t_new below the original bound.
    assert np.all(
        y_new[interp_inputs["t_new"] < np.min(interp_inputs["t"])]
        == interp_inputs["y"][0]
    )


@given(
    timestamps_values_uncertainties_kind(
        extrapolate="below", restrict_fill_value="float"
    )
)
def test_extrapolate_below_with_fill_value_interp1d_unc(interp_inputs):
    # Deal with those cases where at least one of t_new is below the minimum of t and
    # fill_value is a float, which means constant extrapolation with this value.
    y_new = interp1d_unc(**interp_inputs)[1]
    # Check that extrapolation works.
    assert np.all(
        y_new[interp_inputs["t_new"] < np.min(interp_inputs["t"])]
        == interp_inputs["fill_value"]
    )


@given(
    timestamps_values_uncertainties_kind(
        extrapolate="below", restrict_fill_value="tuple"
    )
)
def test_extrapolate_below_with_fill_values_interp1d_unc(interp_inputs):
    # Deal with those cases where at least one of t_new is below the minimum of t and
    # fill_value is a tuple, which means constant extrapolation with its first
    # element.
    y_new = interp1d_unc(**interp_inputs)[1]
    # Check that extrapolation works.
    assert np.all(
        y_new[interp_inputs["t_new"] < np.min(interp_inputs["t"])]
        == interp_inputs["fill_value"][0]
    )


@given(
    timestamps_values_uncertainties_kind(
        sorted_timestamps=True, extrapolate="above", restrict_fill_value="str"
    )
)
def test_extrapolate_above_without_fill_value_interp1d_unc(interp_inputs):
    # Deal with those cases where at least one of t_new is above the maximum of t and
    # fill_value=="extrapolate", which means constant extrapolation from the boundaries.
    y_new = interp1d_unc(**interp_inputs)[1]
    # Check that extrapolation works, meaning in the present case, that the boundary
    # value of y is taken for all t_new above the original bound.
    assert np.all(
        y_new[interp_inputs["t_new"] > np.max(interp_inputs["t"])]
        == interp_inputs["y"][-1]
    )


@given(
    timestamps_values_uncertainties_kind(
        extrapolate="above", restrict_fill_value="float"
    )
)
def test_extrapolate_above_with_fill_value_interp1d_unc(interp_inputs):
    # Deal with those cases where at least one of t_new is above the maximum of t and
    # fill_value is a float, which means constant extrapolation with this value.
    y_new = interp1d_unc(**interp_inputs)[1]
    # Check that extrapolation works.
    assert np.all(
        y_new[interp_inputs["t_new"] > np.max(interp_inputs["t"])]
        == interp_inputs["fill_value"]
    )


@given(
    timestamps_values_uncertainties_kind(
        extrapolate="above", restrict_fill_value="tuple"
    )
)
def test_extrapolate_above_with_fill_values_interp1d_unc(interp_inputs):
    # Deal with those cases where at least one of t_new is above the maximum of t and
    # fill_value is a tuple, which means constant extrapolation with its second element.
    y_new = interp1d_unc(**interp_inputs)[1]
    # Check that extrapolation works.
    assert np.all(
        y_new[interp_inputs["t_new"] > np.max(interp_inputs["t"])]
        == interp_inputs["fill_value"][1]
    )


@given(
    timestamps_values_uncertainties_kind(
        sorted_timestamps=True, extrapolate="below", restrict_fill_unc="str"
    )
)
def test_extrapolate_below_without_fill_unc_interp1d_unc(interp_inputs):
    # Deal with those cases where at least one of t_new is below the minimum of t and
    # fill_unc=="extrapolate", which means constant extrapolation from the boundaries.
    uy_new = interp1d_unc(**interp_inputs)[2]
    # Check that extrapolation works, meaning in the present case, that the boundary
    # value of y is taken for all t_new below the original bound.
    assert np.all(
        uy_new[interp_inputs["t_new"] < np.min(interp_inputs["t"])]
        == interp_inputs["uy"][0]
    )


@given(
    timestamps_values_uncertainties_kind(extrapolate="below", restrict_fill_unc="float")
)
def test_extrapolate_below_with_fill_unc_interp1d_unc(interp_inputs):
    # Deal with those cases where at least one of t_new is below the minimum of t and
    # fill_unc is a float, which means constant extrapolation with this value.
    uy_new = interp1d_unc(**interp_inputs)[2]
    # Check that extrapolation works.
    assert np.all(
        uy_new[interp_inputs["t_new"] < np.min(interp_inputs["t"])]
        == interp_inputs["fill_unc"]
    )


@given(
    timestamps_values_uncertainties_kind(extrapolate="below", restrict_fill_unc="tuple")
)
def test_extrapolate_below_with_fill_uncs_interp1d_unc(interp_inputs):
    # Deal with those cases where at least one of t_new is below the minimum of t and
    # fill_unc is a tuple, which means constant extrapolation with its first element.
    uy_new = interp1d_unc(**interp_inputs)[2]
    # Check that extrapolation works.
    assert np.all(
        uy_new[interp_inputs["t_new"] < np.min(interp_inputs["t"])]
        == interp_inputs["fill_unc"][0]
    )


@given(
    timestamps_values_uncertainties_kind(
        sorted_timestamps=True, extrapolate="above", restrict_fill_unc="str"
    )
)
def test_extrapolate_above_without_fill_unc_interp1d_unc(interp_inputs):
    # Deal with those cases where at least one of t_new is above the maximum of t and
    # fill_unc=="extrapolate", which means constant extrapolation from the boundaries.
    uy_new = interp1d_unc(**interp_inputs)[2]
    # Check that extrapolation works, meaning in the present case, that the boundary
    # value of y is taken for all t_new above the original bound.
    assert np.all(
        uy_new[interp_inputs["t_new"] > np.max(interp_inputs["t"])]
        == interp_inputs["uy"][-1]
    )


@given(
    timestamps_values_uncertainties_kind(extrapolate="above", restrict_fill_unc="float")
)
def test_extrapolate_above_with_fill_unc_interp1d_unc(interp_inputs):
    # Deal with those cases where at least one of t_new is above the maximum of t and
    # fill_unc is a float, which means constant extrapolation with this value.
    uy_new = interp1d_unc(**interp_inputs)[2]
    # Check that extrapolation works.
    assert np.all(
        uy_new[interp_inputs["t_new"] > np.max(interp_inputs["t"])]
        == interp_inputs["fill_unc"]
    )


@given(
    timestamps_values_uncertainties_kind(extrapolate="above", restrict_fill_unc="tuple")
)
def test_extrapolate_above_with_fill_uncs_interp1d_unc(interp_inputs):
    # Deal with those cases where at least one of t_new is above the maximum of t and
    # fill_unc is a tuple, which means constant extrapolation with its second element.
    uy_new = interp1d_unc(**interp_inputs)[2]
    # Check that extrapolation works.
    assert np.all(
        uy_new[interp_inputs["t_new"] > np.max(interp_inputs["t"])]
        == interp_inputs["fill_unc"][1]
    )


@given(timestamps_values_uncertainties_kind(returnC=True, kind_tuple=("linear",)))
def test_compare_returnc_interp1d_unc(interp_inputs):
    # Compare the uncertainties computed from the sensitivities inside the
    # interpolation range and directly.
    uy_new_with_sensitivities = interp1d_unc(**interp_inputs)[2]
    interp_inputs["returnC"] = False
    uy_new_without_sensitivities = interp1d_unc(**interp_inputs)[2]
    # Check that extrapolation results match up to machine epsilon.
    assert_allclose(uy_new_with_sensitivities, uy_new_without_sensitivities, rtol=9e-15)


@given(
    timestamps_values_uncertainties_kind(
        returnC=True, extrapolate=True, kind_tuple=("linear",)
    )
)
def test_failing_returnc_with_extrapolation_interp1d_unc(interp_inputs):
    # Since we have not implemented these cases, for now we
    # check for exception being thrown.
    assume(not isinstance(interp_inputs["fill_unc"], str))
    with raises(NotImplementedError):
        interp1d_unc(**interp_inputs)


@given(
    timestamps_values_uncertainties_kind(
        returnC=True, extrapolate=True, kind_tuple=("linear",), restrict_fill_unc="str"
    )
)
def test_returnc_with_extrapolation_interp1d_unc(interp_inputs):
    # Check if extrapolation with constant values outside interpolation range and
    # returning of sensitivities is callable.
    assert interp1d_unc(**interp_inputs)


@given(
    timestamps_values_uncertainties_kind(
        returnC=True,
        extrapolate=True,
        kind_tuple=("linear",),
        restrict_fill_unc="str",
        sorted_timestamps=True,
    )
)
def test_returnc_with_extrapolation_check_below_bound_interp1d_unc(interp_inputs):
    # Check if extrapolation with constant values outside interpolation range and
    # returning sensitivities work as expected regarding extrapolation values
    # below original bound.
    uy_new, C = interp1d_unc(**interp_inputs)[2:]
    assert np.all(
        uy_new[interp_inputs["t_new"] < np.min(interp_inputs["t"])]
        == interp_inputs["uy"][0]
    )


@given(
    timestamps_values_uncertainties_kind(
        returnC=True,
        extrapolate=True,
        kind_tuple=("linear",),
        restrict_fill_unc="str",
        sorted_timestamps=True,
    )
)
def test_returnc_with_extrapolation_check_uy_new_above_bound_interp1d_unc(
    interp_inputs,
):
    # Check if extrapolation with constant values outside interpolation range and
    # returning sensitivities work as expected regarding extrapolation values
    # above original bound.
    uy_new = interp1d_unc(**interp_inputs)[2]
    assert np.all(
        uy_new[interp_inputs["t_new"] > np.max(interp_inputs["t"])]
        == interp_inputs["uy"][-1]
    )


@given(
    timestamps_values_uncertainties_kind(
        returnC=True, extrapolate=True, kind_tuple=("linear",), restrict_fill_unc="str",
    )
)
def test_returnc_with_extrapolation_check_c_interp1d_unc(interp_inputs,):
    # Check if sensitivity computation parallel to linear interpolation and
    # extrapolation with constant values works as expected regarding the shape and
    # content of the sensitivity matrix.
    C = interp1d_unc(**interp_inputs)[3]

    # Check that C has the right shape.
    assert C.shape == (len(interp_inputs["t_new"]), len(interp_inputs["t"]))

    # Find interpolation range because we reuse it.
    interp_range = (interp_inputs["t_new"] >= np.min(interp_inputs["t"])) | (
        interp_inputs["t_new"] <= np.max(interp_inputs["t"])
    )

    # Check if each row corresponding to an extrapolated value contains exactly one
    # non-zero sensitivity.
    assert np.all(np.count_nonzero(C[np.where(~interp_range)], 1) == 1)

    # Check if each row corresponding to an interpolated value contains either exactly
    # one or exactly two non-zero sensitivities, which are the two possible cases
    # when performing Lagrangian linear interpolation.
    assert np.all(
        np.any(
            (
                np.count_nonzero(C[np.where(interp_range)], 1) == 2,
                np.count_nonzero(C[np.where(interp_range)], 1) == 1,
            ),
            0,
        )
    )

    # Check if each row of sensitivities sum to one, which should hold for the
    # Lagrangians and proves equality with one for extrapolation sensitivities.
    assert_allclose(np.sum(C, 1), np.ones_like(interp_inputs["t_new"]))


@given(
    timestamps_values_uncertainties_kind(
        returnC=True, kind_tuple=("previous", "next", "nearest",)
    )
)
def test_value_error_for_returnc_interp1d_unc(interp_inputs):
    # Check erroneous calls with returnC and wrong kind.
    with raises(NotImplementedError):
        interp1d_unc(**interp_inputs)


@given(st.integers(min_value=3, max_value=1000))
def test_linear_uy_in_interp1d_unc(n,):
    # Check for given input, if interpolated uncertainties equal 1 and
    # :math:`sqrt(2) / 2`.
    dt_unit = 2
    dt_half = dt_unit / 2
    t_new = np.arange(0, n, dt_half)
    t_unit = np.arange(0, n + dt_half, dt_unit)
    y = uy_unit = np.ones_like(t_unit)
    uy_new = interp1d_unc(t_new, t_unit, y, uy_unit, "linear")[2]
    assert np.all(uy_new[0:n:dt_unit] == 1) and np.all(
        uy_new[1:n:dt_unit] == np.sqrt(2) / 2
    )


@given(timestamps_values_uncertainties_kind())
def test_wrong_input_lengths_call_interp1d(interp_inputs):
    # Check erroneous calls with unequally long inputs.
    with raises(ValueError):
        interp_inputs["uy"] = np.tile(interp_inputs["uy"], 2)
        interp1d_unc(**interp_inputs)


@given(timestamps_values_uncertainties_kind(kind_tuple=("spline", "least-squares")))
def test_raise_not_implemented_yet_interp1d(interp_inputs):
    # Check that not implemented versions raise exceptions.
    with raises(NotImplementedError):
        interp1d_unc(**interp_inputs)


@given(timestamps_values_uncertainties_kind(extrapolate=True))
def test_raise_value_error_interp1d_unc(interp_inputs):
    # Check that interpolation with points outside the original domain raises
    # exception if requested.
    interp_inputs["bounds_error"] = True
    with raises(ValueError):
        interp1d_unc(**interp_inputs)



# -*- coding: utf-8 -*-

"""
The :mod:`PyDynamic.uncertainty.propagate_DFT` module implements methods for
the propagation of uncertainties in the application of the DFT, inverse DFT,
deconvolution and multiplication in the frequency domain, transformation from
amplitude and phase to real and imaginary parts and vice versa.

The corresponding scientific publications is
    S. Eichstädt und V. Wilkens
    GUM2DFT — a software tool for uncertainty evaluation of transient signals
    in the frequency domain. *Measurement Science and Technology*, 27(5),
    055001, 2016. [DOI: `10.1088/0957-0233/27/5/055001
    <http://dx.doi.org/10.1088/0957-0233/27/5/055001>`_]

This module contains the following functions:

* :func:`GUM_DFT`: Calculation of the DFT of the time domain signal x and
  propagation of the squared uncertainty Ux associated with the time domain
  sequence x to the real and imaginary parts of the DFT of x
* :func:`GUM_iDFT`: GUM propagation of the squared uncertainty UF associated with
  the DFT values F through the inverse DFT
* :func:`GUM_DFTfreq`: Return the Discrete Fourier Transform sample frequencies
* :func:`DFT_transferfunction`: Calculation of the transfer function H = Y/X in the
  frequency domain with X being the Fourier transform
  of the system's input signal and Y that of the output signal
* :func:`DFT_deconv`: Deconvolution in the frequency domain
* :func:`DFT_multiply`: Multiplication in the frequency domain
* :func:`AmpPhase2DFT`: Transformation from magnitude and phase to real and
  imaginary parts
* :func:`DFT2AmpPhase`: Transformation from real and imaginary parts to magnitude
  and phase
* :func:`AmpPhase2Time`: Transformation from amplitude and phase to time domain
* :func:`Time2AmpPhase`: Transformation from time domain to amplitude and phase
"""

import warnings
from typing import Dict, Optional, Tuple, Union

import numpy as np
from scipy import sparse

__all__ = [
    "GUM_DFT",
    "GUM_iDFT",
    "GUM_DFTfreq",
    "DFT_transferfunction",
    "DFT_deconv",
    "DFT_multiply",
    "AmpPhase2DFT",
    "DFT2AmpPhase",
    "AmpPhase2Time",
    "Time2AmpPhase",
    "Time2AmpPhase_multi",
]


def _apply_window(
    x: np.ndarray, Ux: Union[np.ndarray, float], window: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """Apply time domain window to signal x of equal length and propagate uncertainties

    Parameters
    ----------
    x : np.ndarray of shape (N, )
        vector of time domain signal values
    Ux : np.ndarray of shape (N, N) or float
        covariance matrix associated with x or noise variance as float
    window : np.ndarray of shape (N, )
        vector of time domain window (same length as x)

    Returns
    -------
    xw, Uxw : np.ndarray of shape (N, ) and (N, N)
        transformed signal and associated uncertainties

    Raises
    ------
    AssertionError
        If the dimensions of x, Ux and window do not match or Ux is not square.
    """
    assert len(x) == len(window)
    if not isinstance(Ux, float):
        assert Ux.shape[0] == Ux.shape[1] and Ux.shape[0] == len(x)
    xw = x.copy() * window
    if isinstance(Ux, float):
        Uxw = Ux * window ** 2
    else:
        Uxw = _prod(window, _prod(Ux, window))
    return xw, Uxw


def _prod(A: np.ndarray, B: np.ndarray) -> np.ndarray:
    """Calculate the product of a matrix' diagonal and a vector

    Calculate the product that corresponds to diag(A)*B or A*diag(B),
    respectively; depending on which of A,B is the matrix and which the vector.

    Parameters
    ----------
    A, B : np.ndarray of shape (N, )
        one is a vector and the other a matrix of which we want to multipy the diagonal
        with the vector

    Returns
    -------
    np.ndarray
        The product of the matrix' diagonal and the vector

    Raises
    ------
    ValueError
        If the dimensions of A and B do not match.
    """
    if len(A.shape) == 1 and len(B.shape) == 2:  # A is the vector and B the matrix
        C = np.zeros_like(B)
        for k in range(C.shape[0]):
            C[k, :] = A[k] * B[k, :]
        return C
    elif len(A.shape) == 2 and len(B.shape) == 1:  # A is the matrix and B the vector
        C = np.zeros_like(A)
        for k in range(C.shape[1]):
            C[:, k] = A[:, k] * B[k]
        return C
    else:
        raise ValueError("_prod: Wrong dimension of inputs")


def _matprod(
    M: np.ndarray, V: np.ndarray, W: np.ndarray, return_as_matrix: Optional[bool] = True
) -> np.ndarray:
    """Calculate the matrix-matrix-matrix product (V1,V2)M(W1,W2)

    Calculate the product for V=(V1,V2) and W=(W1,W2).

    Parameters
    ----------
    M : np.ndarray
        M can be sparse, one-dimensional or a full (quadratic) matrix.
    V, W : np.ndarray
        V=(V1,V2) and W=(W1,W2)

    Returns
    -------
    np.ndarray
        matrix product (V1,V2)M(W1,W2)

    Raises
    ------
    AssertionError
        If the dimensions of M, V and W do not match or M is a non-square matrix.
    """
    if len(M.shape) == 2:
        assert M.shape[0] == M.shape[1]
    assert M.shape[0] == V.shape[0]
    assert V.shape == W.shape
    N = V.shape[0] // 2
    v1 = V[:N]
    v2 = V[N:]
    w1 = W[:N]
    w2 = W[N:]
    if isinstance(M, sparse.dia_matrix):
        nrows = M.shape[0]
        offset = M.offsets
        diags = M.data
        A = diags[0][:N]
        B = diags[1][offset[1] : nrows + offset[1]]
        D = diags[0][N:]
        return np.diag(v1 * A * w1 + v2 * B * w1 + v1 * B * w2 + v2 * D * w2)
    elif len(M.shape) == 1:
        A = M[:N]
        D = M[N:]
        if return_as_matrix:
            return np.diag(v1 * A * w1 + v2 * D * w2)
        else:
            return np.r_[v1 * A * w1 + v2 * D * w2]
    else:
        A = M[:N, :N]
        B = M[:N, N:]
        D = M[N:, N:]
        return (
            _prod(v1, _prod(A, w1))
            + _prod(v2, _prod(B.T, w1))
            + _prod(v1, _prod(B, w2))
            + _prod(v2, _prod(D, w2))
        )


def GUM_DFT(
    x: np.ndarray,
    Ux: Union[np.ndarray, float],
    N: Optional[int] = None,
    window: Optional[np.ndarray] = None,
    CxCos: Optional[np.ndarray] = None,
    CxSin: Optional[np.ndarray] = None,
    returnC: Optional[bool] = False,
    mask: Optional[np.ndarray] = None,
) -> Union[
    Tuple[np.ndarray, Union[Tuple[np.ndarray, np.ndarray, np.ndarray], np.ndarray]],
    Tuple[
        np.ndarray,
        Union[Tuple[np.ndarray, np.ndarray, np.ndarray], np.ndarray],
        Dict[str, np.ndarray],
    ],
]:
    """Calculation of the DFT with propagation of uncertainty

    Calculation of the DFT of the time domain signal x and propagation of
    the squared uncertainty Ux associated with the time domain sequence x to
    the real and imaginary parts of the DFT of x.

    Parameters
    ----------
    x : np.ndarray of shape (M,)
        vector of time domain signal values
    Ux : np.ndarray of shape (M,) or of shape (M,M) or float
        covariance matrix associated with x, or vector of squared standard
        uncertainties, or noise variance as float
    N : int, optional
        length of time domain signal for DFT; N>=len(x)
    window : np.ndarray of shape (M,), optional
        vector of the time domain window values
    CxCos : np.ndarray, optional
        cosine part of sensitivity matrix
    CxSin : np.ndarray, optional
        sine part of sensitivity matrix
    returnC : bool, optional
        if True, return sensitivity matrix blocks, if False (default) do not return them
    mask: ndarray of dtype bool, optional
        calculate DFT values and uncertainties only at those frequencies
        where mask is True

    Returns
    -------
    F : np.ndarray
        vector of complex valued DFT values or of its real and imaginary parts
    UF : np.ndarray
        covariance matrix associated with real and imaginary part of F
    CxCos and CxSin : Dict
        Keys are "CxCos", "CxSin" and values the respective sensitivity matrix entries

    References
    ----------
        * Eichstädt and Wilkens [Eichst2016]_

    Raises
    ------
    ValueError
        If N < len(x)
    """
    L = 0
    # Apply the chosen window for the application of the FFT.
    if isinstance(window, np.ndarray):
        x, Ux = _apply_window(x, Ux, window)
    if isinstance(N, int):
        L = N - len(x)
        if L < 0:
            raise ValueError(
                "N needs to be greater or equal than the length of x, "
                f"but N = {N} and len(x) = {len(x)} were given."
            )
        x = np.r_[
            x.copy(),
            np.zeros(
                L,
            ),
        ]  # zero-padding
    N = len(x)
    if np.mod(N, 2) == 0:  # N is even
        M = N + 2
    else:
        M = N + 1

    if isinstance(mask, np.ndarray):
        F = np.fft.rfft(x)[mask]  # calculation of best estimate
        # In real, imag format in accordance with GUM S2
        F = np.r_[np.real(F), np.imag(F)]
        warnings.warn(
            "In a future release, because of issues with the current version, "
            "\nthe handling of masked DFT arrays will be changed to use numpy "
            "masked arrays.",
            DeprecationWarning,
        )
    else:
        F = np.fft.rfft(x)  # calculation of best estimate
        # In real, imag format in accordance with GUM S2
        F = np.r_[np.real(F), np.imag(F)]
        mask = np.ones(len(F) // 2, dtype=bool)
    Nm = 2 * np.sum(mask)

    # For simplified calculation of sensitivities
    beta = 2 * np.pi * np.arange(N - L) / N

    # sensitivity matrix wrt cosine part
    Cxkc = lambda k: np.cos(k * beta)[np.newaxis, :]
    # sensitivity matrix wrt sinus part
    Cxks = lambda k: -np.sin(k * beta)[np.newaxis, :]

    if isinstance(Ux, float):
        UF = np.zeros(Nm)
        km = 0
        for k in range(M // 2):  # Block cos/cos
            if mask[k]:
                UF[km] = np.sum(Ux * Cxkc(k) ** 2)
                km += 1
        km = 0
        for k in range(M // 2):  # Block sin/sin
            if mask[k]:
                UF[Nm // 2 + km] = np.sum(Ux * Cxks(k) ** 2)
                km += 1
    else:  # general method
        if len(Ux.shape) == 1:
            Ux = np.diag(Ux)
        if not isinstance(CxCos, np.ndarray):
            CxCos = np.zeros((Nm // 2, N - L))
            CxSin = np.zeros((Nm // 2, N - L))
            km = 0
            for k in range(M // 2):
                if mask[k]:
                    CxCos[km, :] = Cxkc(k)
                    CxSin[km, :] = Cxks(k)
                    km += 1
        UFCC = np.dot(CxCos, np.dot(Ux, CxCos.T))
        UFCS = np.dot(CxCos, np.dot(Ux, CxSin.T))
        UFSS = np.dot(CxSin, np.dot(Ux, CxSin.T))
        try:
            UF = np.vstack(
                (np.hstack((UFCC, UFCS)), np.hstack((UFCS.T, UFSS)))
            )  # type: Union[Tuple[np.ndarray, np.ndarray, np.ndarray], np.ndarray]
        except MemoryError:
            print(
                "Could not put covariance matrix together due to memory constraints."
            )
            print(
                "Returning the three blocks (A,B,C) such that U = [[A,B],"
                "[B.T,C]] instead."
            )
            # Return blocks only because of lack of memory.
            UF = (UFCC, UFCS, UFSS)  # type:
            # Union[Tuple[np.ndarray, np.ndarray, np.ndarray], np.ndarray]

    if returnC:
        # Return sensitivities if requested.
        return F, UF, {"CxCos": CxCos, "CxSin": CxSin}
    else:
        return F, UF


def GUM_iDFT(
    F: np.ndarray,
    UF: np.ndarray,
    Nx: Optional[int] = None,
    Cc: Optional[np.ndarray] = None,
    Cs: Optional[np.ndarray] = None,
    returnC: Optional[bool] = False,
) -> Union[
    Tuple[np.ndarray, np.ndarray],
    Tuple[
        np.ndarray,
        np.ndarray,
        Dict[str, np.ndarray],
    ],
]:
    """Propagation of squared uncertainties UF associated with the DFT values F

    GUM propagation of the squared uncertainties UF associated with the DFT
    values F through the inverse DFT.

    The matrix UF is assumed to be for real and imaginary part with blocks:
    UF = [[u(R,R), u(R,I)],[u(I,R),u(I,I)]]
    and real and imaginary part obtained from calling rfft (DFT for
    real-valued signal)

    Parameters
    ----------
    F : np.ndarray of shape (2M,)
        vector of real and imaginary parts of a DFT result
    UF: np.ndarray of shape (2M,2M)
        covariance matrix associated with real and imaginary parts of F
    Nx: int, optional
        number of samples of iDFT result
    Cc: np.ndarray, optional
        cosine part of sensitivities (without scaling factor 1/N)
    Cs: np.ndarray, optional
        sine part of sensitivities (without scaling factor 1/N)
    returnC : bool, optional
        If True, return sensitivity matrix blocks (without scaling factor 1/N),
        if False do not return them

    Returns
    -------
    x : np.ndarray
        vector of time domain signal values
    Ux : np.ndarray
        covariance matrix associated with x
    Cc and Cs : Dict
        Keys are "Cc", "Cs" and values the respective sensitivity matrix entries

    References
    ----------
        * Eichstädt and Wilkens [Eichst2016]_

    Raises
    ------
    ValueError
        If Nx is not smaller than dimension of UF - 2
    """
    N = UF.shape[0] - 2

    if Nx is None:
        Nx = N
    elif Nx > UF.shape[0] - 2:
        raise ValueError(
            f"GUM_iDFT: Nx if provided is expected to be smaller or equal to number "
            f"of rows and columns of UF - 2, but UF is of shape {UF.shape} and Nx ="
            f" {Nx}."
        )

    beta = 2 * np.pi * np.arange(Nx) / N

    # calculate inverse DFT; Note: scaling factor 1/N is accounted for at the end
    x = np.fft.irfft(F[: N // 2 + 1] + 1j * F[N // 2 + 1 :])[:Nx]
    if not isinstance(Cc, np.ndarray):  # calculate sensitivities
        Cc = np.zeros((Nx, N // 2 + 1))
        Cc[:, 0] = 1.0
        Cc[:, -1] = np.cos(np.pi * np.arange(Nx))
        for k in range(1, N // 2):
            Cc[:, k] = 2 * np.cos(k * beta)

    if not isinstance(Cs, np.ndarray):
        Cs = np.zeros((Nx, N // 2 + 1))
        Cs[:, 0] = 0.0
        Cs[:, -1] = -np.sin(np.pi * np.arange(Nx))
        for k in range(1, N // 2):
            Cs[:, k] = -2 * np.sin(k * beta)

    # calculate blocks of uncertainty matrix
    if len(UF.shape) == 2:
        RR = UF[: N // 2 + 1, : N // 2 + 1]
        RI = UF[: N // 2 + 1, N // 2 + 1 :]
        II = UF[N // 2 + 1 :, N // 2 + 1 :]
        # propagate uncertainties
        Ux = np.dot(Cc, np.dot(RR, Cc.T))
        Ux = Ux + 2 * np.dot(Cc, np.dot(RI, Cs.T))
        Ux = Ux + np.dot(Cs, np.dot(II, Cs.T))
    else:
        RR = UF[: N // 2 + 1]
        II = UF[N // 2 + 1 :]
        Ux = np.dot(Cc, _prod(RR, Cc.T)) + np.dot(Cs, _prod(II, Cs.T))

    if returnC:
        return x, Ux / N ** 2, {"Cc": Cc, "Cs": Cs}
    else:
        return x, Ux / N ** 2


def GUM_DFTfreq(N, dt=1):
    """Return the Discrete Fourier Transform sample frequencies

    Parameters
    ----------
        N: int
            window length
        dt: float
            sample spacing (inverse of sampling rate)

    Returns
    -------
        f: ndarray
            Array of length ``n//2 + 1`` containing the sample frequencies

    See also
    --------
        `mod`::numpy.fft.rfftfreq

    """

    return np.fft.rfftfreq(N, dt)


def DFT2AmpPhase(
    F: np.ndarray,
    UF: np.ndarray,
    keep_sparse: Optional[bool] = False,
    tol: Optional[float] = 1.0,
    return_type: Optional[str] = "separate",
) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray, np.ndarray]]:
    """Transformation from real and imaginary parts to magnitude and phase

    Calculate the matrix
    U_AP = [[U1,U2],[U2^T,U3]]
    associated with magnitude and phase of the vector F=[real,imag]
    with associated covariance matrix U_F=[[URR,URI],[URI^T,UII]]

    Parameters
    ----------
    F : np.ndarray of shape (2M,)
        vector of real and imaginary parts of a DFT result
    UF : np.ndarray of shape (2M,2M)
        covariance matrix associated with F
    keep_sparse : bool, optional
        if true then UAP will be sparse if UF is one-dimensional
    tol : float, optional
        lower bound for A/uF below which a warning will be issued
        concerning unreliable results
    return_type : str, optional
        If "separate" then magnitude and phase are returned as separate
        arrays A and P. Otherwise the list [A, P] is returned

    If ``return_type == separate``:

    Returns
    -------
    A : np.ndarray
        vector of magnitude values
    P : np.ndarray
        vector of phase values in radians, in the range [-pi, pi], but only
        present if ``return_type = 'separate'``
    UAP : np.ndarray
        covariance matrix associated with (A,P)

    Otherwise:

    Returns
    -------
    AP : np.ndarray
        vector of magnitude and phase values
    UAP : np.ndarray
        covariance matrix associated with AP
    """
    # calculate inverse DFT
    N = len(F) - 2
    R = F[: N // 2 + 1]
    I = F[N // 2 + 1 :]

    A = np.sqrt(R ** 2 + I ** 2)  # absolute value
    P = np.arctan2(I, R)  # phase value
    if len(UF.shape) == 1:
        uF = 0.5 * (
            np.sqrt(UF[: N // 2 + 1]) + np.sqrt(UF[N // 2 + 1 :])
        )  # uncertainty of real,imag
    else:
        uF = 0.5 * (
            np.sqrt(np.diag(UF[: N // 2 + 1, : N // 2 + 1]))
            + np.sqrt(np.diag(UF[N // 2 + 1 :, N // 2 + 1 :]))
        )
    if np.any(A / uF < tol):
        print(
            "DFT2AmpPhase Warning\n Some amplitude values are below the "
            "defined threshold.\nThe GUM formulas may become unreliable and a Monte "
            "Carlo approach is recommended instead.\nThe actual minimum value of A/uF"
            f"is {round((A / uF).min(), 2)} and the threshold is {round(tol, 2)}."
        )
    aR = R / A  # sensitivities
    aI = I / A
    pR = -I / A ** 2
    pI = R / A ** 2

    if len(UF.shape) == 1:  # uncertainty calculation of zero correlation
        URR = UF[: N // 2 + 1]
        UII = UF[N // 2 + 1 :]
        U11 = URR * aR ** 2 + UII * aI ** 2
        U12 = aR * URR * pR + aI * UII * pI
        U22 = URR * pR ** 2 + UII * pI ** 2
        UAP = sparse.diags([np.r_[U11, U22], U12, U12], [0, N // 2 + 1, -(N // 2 + 1)])
        if not keep_sparse:
            UAP = UAP.toarray()
    else:  # uncertainty calculation for full covariance
        URR = UF[: N // 2 + 1, : N // 2 + 1]
        URI = UF[: N // 2 + 1, N // 2 + 1 :]
        UII = UF[N // 2 + 1 :, N // 2 + 1 :]
        U11 = (
            _prod(aR, _prod(URR, aR))
            + _prod(aR, _prod(URI, aI))
            + _prod(aI, _prod(URI.T, aR))
            + _prod(aI, _prod(UII, aI))
        )
        U12 = (
            _prod(aR, _prod(URR, pR))
            + _prod(aR, _prod(URI, pI))
            + _prod(aI, _prod(URI.T, pR))
            + _prod(aI, _prod(UII, pI))
        )
        U22 = (
            _prod(pR, _prod(URR, pR))
            + _prod(pR, _prod(URI, pI))
            + _prod(pI, _prod(URI.T, pR))
            + _prod(pI, _prod(UII, pI))
        )
        UAP = np.vstack((np.hstack((U11, U12)), np.hstack((U12.T, U22))))

    if return_type == "separate":
        return A, P, UAP  # amplitude and phase as separate variables
    else:
        return np.r_[A, P], UAP


def AmpPhase2DFT(
    A: np.ndarray, P: np.ndarray, UAP: np.ndarray, keep_sparse: Optional[bool] = False
) -> Tuple[np.ndarray, np.ndarray]:
    """Transformation from magnitude and phase to real and imaginary parts

    Calculate the vector F=[real,imag] and propagate the covariance matrix UAP
    associated with [A, P]

    Parameters
    ----------
    A : np.ndarray of shape (N,)
        vector of magnitude values
    P : np.ndarray of shape (N,)
        vector of phase values (in radians)
    UAP : np.ndarray of shape (2N,2N) or of shape (2N,)
        covariance matrix associated with (A,P) or vector of squared standard
        uncertainties [u^2(A),u^2(P)]
    keep_sparse : bool, optional
        whether to transform sparse matrix to numpy array or not

    Returns
    -------
    F : np.ndarray of shape (2N,)
        vector of real and imaginary parts of DFT result
    UF : np.ndarray of shape (2N,2N)
        covariance matrix associated with F

    Raises
    ------
    ValueError
        If dimensions of A, P and UAP do not match.
    """
    if len(A.shape) != 1:
        raise ValueError(
            f"AmpPhase2DFT: A is expected to be a 1-dimensional vector but is of shape"
            f" {A.shape}."
        )
    if A.shape != P.shape:
        raise ValueError(
            f"AmpPhase2DFT: The dimensions of A and P are expected to match but A is "
            f"of shape {A.shape} and P is of shape {P.shape}."
        )
    if UAP.shape != (2 * len(A), 2 * len(A)) and UAP.shape != (2 * len(A),):
        raise ValueError(
            f"AmpPhase2DFT: Each dimension of UAP is expected to be twice the length "
            f"of A, but A is of length {len(A)} and UAP is of shape {UAP.shape}."
        )

    # calculation of F
    F = np.r_[A * np.cos(P), A * np.sin(P)]  # calculation of best estimate

    # calculation of sensitivities
    CRA = np.cos(P)
    CRP = -A * np.sin(P)
    CIA = np.sin(P)
    CIP = A * np.cos(P)

    # assignment of uncertainty blocks in UAP
    N = len(A)
    if UAP.shape == (2 * N,):  # zero correlation; just standard deviations
        Ua = UAP[:N]
        Up = UAP[N:]
        U11 = CRA * Ua * CRA + CRP * Up * CRP
        U12 = CRA * Ua * CIA + CRP * Up * CIP
        U22 = CIA * Ua * CIA + CIP * Up * CIP
        UF = sparse.diags([np.r_[U11, U22], U12, U12], [0, N, -N])
        if not keep_sparse:
            UF = UF.toarray()
    else:
        if isinstance(UAP, sparse.dia_matrix):
            nrows = 2 * N
            offset = UAP.offsets
            diags = UAP.data
            Uaa = diags[0][:N]
            Uap = diags[1][offset[1] : nrows + offset[1]]
            Upp = diags[0][N:]

            U11 = Uaa * CRA ** 2 + CRP * Uap * CRA + CRA * Uap * CRP + Upp * CRP ** 2
            U12 = CRA * Uaa * CIA + CRP * Uap * CIA + CRA * Uap * CIA + CRP * Upp * CIP
            U22 = Uaa * CIA ** 2 + CIP * Uap * CIA + CIA * Uap * CIP + Upp * CIP ** 2

            UF = sparse.diags(
                [np.r_[U11, U22], U12, U12], [0, N, -N]
            )  # default is sparse
            if not keep_sparse:
                UF = UF.toarray()  # fall back to non-sparse
        else:
            Uaa = UAP[:N, :N]
            Uap = UAP[:N, N:]
            Upp = UAP[N:, N:]

            U11 = (
                _prod(CRA, _prod(Uaa, CRA))
                + _prod(CRP, _prod(Uap.T, CRA))
                + _prod(CRA, _prod(Uap, CRP))
                + _prod(CRP, _prod(Upp, CRP))
            )
            U12 = (
                _prod(CRA, _prod(Uaa, CIA))
                + _prod(CRP, _prod(Uap.T, CIA))
                + _prod(CRA, _prod(Uap, CIP))
                + _prod(CRP, _prod(Upp, CIP))
            )
            U22 = (
                _prod(CIA, _prod(Uaa, CIA))
                + _prod(CIP, _prod(Uap.T, CIA))
                + _prod(CIA, _prod(Uap, CIP))
                + _prod(CIP, _prod(Upp, CIP))
            )

            # stack together the full covariance matrix
            UF = np.vstack((np.hstack((U11, U12)), np.hstack((U12.T, U22))))

    return F, UF


def Time2AmpPhase(
    x: np.ndarray, Ux: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transformation from time domain to amplitude and phase via DFT

    Parameters
    ----------
    x : np.ndarray of shape (N,)
        time domain signal
    Ux : np.ndarray of shape (N,N)
        squared uncertainty associated with x

    Returns
    -------
    A : np.ndarray
        amplitude values
    P : np.ndarray
        phase values
    UAP : np.ndarray
        covariance matrix associated with [A,P]
    """
    F, UF = GUM_DFT(x, Ux)  # propagate to DFT domain
    A, P, UAP = DFT2AmpPhase(F=F, UF=UF, return_type="separate")  # propagate to
    # amplitude and phase
    return A, P, UAP


def Time2AmpPhase_multi(x, Ux, selector=None):
    """Transformation from time domain to amplitude and phase

    Perform transformation for a set of M signals of the same type.

    Parameters
    ----------
        x: np.ndarray of shape (M, nx)
            M time domain signals of length nx
        Ux: np.ndarray of shape (M,)
            squared standard deviations representing noise variances of the
            signals x
        selector: np.ndarray of shape (L,), optional
            indices of amplitude and phase values that should be returned;
            default is 0:N-1
    Returns
    -------
        A: np.ndarray of shape (M,N)
            amplitude values
        P: np.ndarray of shape (M,N)
            phase values
        UAP: np.ndarray of shape (M, 3N)
            diagonals of the covariance matrices: [diag(UAA), diag(UAP),
            diag(UPP)]
    """
    M, nx = x.shape
    assert len(Ux) == M
    N = nx // 2 + 1
    if not isinstance(selector, np.ndarray):
        selector = np.arange(nx // 2 + 1)
    ns = len(selector)

    A = np.zeros((M, ns))
    P = np.zeros_like(A)
    UAP = np.zeros((M, 3 * ns))
    CxCos = None
    CxSin = None
    for m in range(M):
        F, UF, CX = GUM_DFT(x[m, :], Ux[m], CxCos, CxSin, returnC=True)
        CxCos = CX["CxCos"]
        CxSin = CX["CxSin"]
        A_m, P_m, UAP_m = DFT2AmpPhase(F, UF, keep_sparse=True)
        A[m, :] = A_m[selector]
        P[m, :] = P_m[selector]
        UAP[m, :ns] = UAP_m.data[0][:N][selector]
        UAP[m, ns : 2 * ns] = UAP_m.data[1][
            UAP_m.offsets[1] : 2 * N + UAP_m.offsets[1]
        ][selector]
        UAP[m, 2 * ns :] = UAP_m.data[0][N:][selector]

    return A, P, UAP


def AmpPhase2Time(
    A: np.ndarray, P: np.ndarray, UAP: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """Transformation from amplitude and phase to time domain

    GUM propagation of covariance matrix UAP associated with DFT amplitude A and
    phase P to the result of the inverse DFT. Uncertainty UAP is assumed to be given
    for amplitude and phase with blocks:
    UAP = [[u(A,A), u(A,P)],[u(P,A),u(P,P)]]

    Parameters
    ----------
    A : np.ndarray of shape (N, )
        vector of amplitude values
    P : np.ndarray of shape (N, )
        vector of phase values (in rad)
    UAP : np.ndarray of shape (2N, 2N)
        covariance matrix associated with [A,P]

    Returns
    -------
    x : np.ndarray of shape (N, )
        vector of time domain values
    Ux : np.ndarray of shape (2N, 2N)
        covariance matrix associated with x

    Raises
    ------
    ValueError
        If dimension of UAP is not even.
    """
    N = UAP.shape[0] - 2
    if A.shape != P.shape:
        raise ValueError(
            f"AmpPhase2Time: The dimensions of A and P are expected to match but A is "
            f"of shape {A.shape} and P is of shape {P.shape}."
        )
    if np.mod(N, 2) != 0:
        raise ValueError(
            "AmpPhase2Time: Dimension of covariance matrix UAP is "
            f"supposed to be even, but UAP has {N+2} rows. Please "
            f"provide covariance matrix of dimension {len(A) + len(P)} x "
            f" {len(A) + len(P)}."
        )
    beta = 2 * np.pi * np.arange(N) / N

    # calculate inverse DFT
    F = A * np.exp(1j * P)
    x = np.fft.irfft(F)

    Pf = np.r_[P, -P[-2:0:-1]]  # phase values to take into account symmetric part
    Cc = np.zeros((N, N // 2 + 1))  # sensitivities wrt cosine part
    Cc[:, 0] = np.cos(P[0])
    Cc[:, -1] = np.cos(P[-1] + np.pi * np.arange(N))
    for k in range(1, N // 2):
        Cc[:, k] = 2 * np.cos(Pf[k] + k * beta)

    Cs = np.zeros((N, N // 2 + 1))  # sensitivities wrt sinus part
    Cs[:, 0] = -A[0] * np.sin(P[0])
    Cs[:, -1] = -A[-1] * np.sin(P[-1] + np.pi * np.arange(N))
    for k in range(1, N // 2):
        Cs[:, k] = -A[k] * 2 * np.sin(Pf[k] + k * beta)

    # calculate blocks of uncertainty matrix
    if len(UAP.shape) == 1:
        AA = UAP[: N // 2 + 1]
        PP = UAP[N // 2 + 1 :]
        Ux = np.dot(Cc, _prod(AA, Cc.T)) + np.dot(Cs, _prod(PP, Cs.T))
    else:
        if isinstance(UAP, sparse.dia_matrix):
            nrows = UAP.shape[0]
            n = nrows // 2
            offset = UAP.offsets
            diags = UAP.data
            AA = diags[0][:n]
            AP = diags[1][offset[1] : nrows + offset[1]]
            PP = diags[0][n:]
            Ux = (
                np.dot(Cc, _prod(AA, Cc.T))
                + 2 * np.dot(Cc, _prod(AP, Cs.T))
                + np.dot(Cs, _prod(PP, Cs.T))
            )
        else:
            AA = UAP[: N // 2 + 1, : N // 2 + 1]
            AP = UAP[: N // 2 + 1, N // 2 + 1 :]
            PP = UAP[N // 2 + 1 :, N // 2 + 1 :]
            # propagate uncertainties
            Ux = (
                np.dot(Cc, np.dot(AA, Cc.T))
                + 2 * np.dot(Cc, np.dot(AP, Cs.T))
                + np.dot(Cs, np.dot(PP, Cs.T))
            )

    return x, Ux / N ** 2


# for backward compatibility
GUMdeconv = lambda H, Y, UH, UY: DFT_deconv(H, Y, UH, UY)


def DFT_transferfunction(X, Y, UX, UY):
    """Calculation of the transfer function H = Y/X in the frequency domain

    Calculate the transfer function with X being the Fourier transform
    of the system's input signal and Y that of the output signal.

    Parameters
    ----------
        X: np.ndarray
            real and imaginary parts of the system's input signal
        Y: np.ndarray
            real and imaginary parts of the system's output signal
        UX: np.ndarray
            covariance matrix associated with X
        UY: np.ndarray
            covariance matrix associated with Y

    Returns
    -------
        H: np.ndarray
            real and imaginary parts of the system's frequency response
        UH: np.ndarray
            covariance matrix associated with H

    This function only calls `DFT_deconv`.
    """
    return DFT_deconv(X, Y, UX, UY)


def DFT_deconv(
    H: np.ndarray, Y: np.ndarray, UH: np.ndarray, UY: np.ndarray
) -> Tuple[np.ndarray, Union[Tuple[np.ndarray, np.ndarray, np.ndarray], np.ndarray]]:
    """Deconvolution in the frequency domain

    GUM propagation of uncertainties for the deconvolution X = Y/H with Y and
    H being the Fourier transform of the measured signal and of the system's impulse
    response, respectively.

    This function returns the covariance matrix as a tuple of blocks if too large for
    complete storage in memory.

    Parameters
    ----------
    H : np.ndarray of shape (2M,)
        real and imaginary parts of frequency response values (M an even integer)
    Y : np.ndarray of shape (2M,)
        real and imaginary parts of DFT values
    UH : np.ndarray of shape (2M,2M) or (2M,)
        full covariance or diagonal of the covariance matrix associated with H
    UY : np.ndarray of shape (2M,2M) or (2M,)
        full covariance or diagonal of the covariance matrix associated with Y

    Returns
    -------
    X : np.ndarray of shape (2M,)
        real and imaginary parts of DFT values of deconv result
    UX : np.ndarray of shape (2M,2M) or 3-tuple of np.ndarray of shape (M,M)
        Covariance matrix associated with real and imaginary part of X. If the matrix
        fully assembled does not fit the memory, we return the auto-covariance for the
        real parts ``URRX``and the imaginary parts ``UIIX`` and the covariance between
        the real and imaginary parts ``URIX`` as separate
        :class:`np.ndarrays <numpy.ndarray>` arranged as follows: ``(URRX, URIX, UIIX)``

    References
    ----------
        * Eichstädt and Wilkens [Eichst2016]_

    Raises
    ------
    ValueError
        If dimensions of H, Y, UY and UH do not match accordingly.
    """
    if len(H) != len(Y):
        raise ValueError(
            f"The dimensions of H and Y are expected to match but H is of length "
            f"{len(H)} and Y is of length {len(Y)}."
        )

    N = len(UH) - 2
    if np.mod(N, 2) != 0:
        raise ValueError(
            "DFT_deconv: Dimension of covariance matrix UH is "
            f"supposed to be even, but UH has {N+2} rows. Please "
            f"provide covariance matrix of dimension {len(H)} x "
            f" {len(H)}."
        )

    if UY.shape not in ((len(Y), len(Y)), (len(Y),)):
        raise ValueError(
            f"DFT_deconv:Each dimension of UY is expected to match the length of "
            f"Y, but Y is of length {len(Y)} and UY is of shape {UY.shape}."
        )

    if UH.shape not in ((len(H), len(H)), (len(H),)):
        raise ValueError(
            f"DFT_deconv: The dimensions of UH and H are expected to match, "
            f"but H is of length {len(H)} and UH is of shape {UH.shape}."
        )

    # real and imaginary parts of system and signal
    rH, iH = H[: N // 2 + 1], H[N // 2 + 1 :]
    rY, iY = Y[: N // 2 + 1], Y[N // 2 + 1 :]

    Yc = Y[: N // 2 + 1] + 1j * Y[N // 2 + 1 :]
    Hc = H[: N // 2 + 1] + 1j * H[N // 2 + 1 :]
    X = np.r_[np.real(Yc / Hc), np.imag(Yc / Hc)]  # type: np.ndarray

    # sensitivities
    norm = rH ** 2 + iH ** 2
    RY = np.r_[rH / norm, iH / norm]
    IY = np.r_[-iH / norm, rH / norm]
    RH = np.r_[
        (-rY * rH ** 2 + rY * iH ** 2 - 2 * iY * iH * rH) / norm ** 2,
        (iY * rH ** 2 - iH * iH ** 2 - 2 * rY * rH * iH) / norm ** 2,
    ]
    IH = np.r_[
        (-iY * rH ** 2 + iY * iH ** 2 + 2 * rY * iH * rH) / norm ** 2,
        (-rY * rH ** 2 + rY * iH ** 2 - 2 * iY * rH * iH) / norm ** 2,
    ]
    # calculate blocks of uncertainty matrix
    URRX = _matprod(UY, RY, RY) + _matprod(UH, RH, RH)
    URIX = _matprod(UY, RY, IY) + _matprod(UH, RH, IH)
    UIIX = _matprod(UY, IY, IY) + _matprod(UH, IH, IH)

    try:
        UX = np.vstack((np.hstack((URRX, URIX)), np.hstack((URIX.T, UIIX))))  # type:
        # np.ndarray
    except MemoryError:
        print(
            "DFT_deconv: Could not put covariance matrix together due to memory "
            "constraints.\nReturning the three blocks (A,B,C) such that U = [[A,B],"
            "[B.T,C]] instead."
        )
        UX = (URRX, URIX, UIIX)
        # type: Union[Tuple[np.ndarray, np.ndarray, np.ndarray], np.ndarray]

    return X, UX


def DFT_multiply(
    Y: np.ndarray, F: np.ndarray, UY: np.ndarray, UF: Optional[np.ndarray] = None
) -> Tuple[np.ndarray, np.ndarray]:
    """Multiplication in the frequency domain

    GUM uncertainty propagation for multiplication in the frequency domain,
    where the second factor F may have an associated uncertainty. This method can be
    used, for instance, for the application of a low-pass filter in the frequency
    domain or the application of deconvolution as a multiplication with an inverse of
    known uncertainty.

    Parameters
    ----------
    Y : np.ndarray of shape (2M,)
        real and imaginary parts of the first factor
    F : np.ndarray of shape (2M,)
        real and imaginary parts of the second factor
    UY : np.ndarray either of shape (2M,) or of shape (2M,2M)
        covariance matrix or squared uncertainty associated with Y
    UF : np.ndarray of shape (2M,2M), optional
        covariance matrix associated with F

    Returns
    -------
    YF : np.ndarray of shape (2M,)
        the product of Y and F
    UYF : np.ndarray of shape (2M,2M)
        the uncertainty associated with YF

    Raises
    ------
    ValueError
        If dimensions of Y and F do not match.
    """
    if len(Y) != len(F):
        raise ValueError(
            f"GUM_multiply: The dimensions of Y and F are expected to match but Y is "
            f"of length {len(Y)} and F is of length {len(F)}."
        )

    def _calcU(A, UB):
        # uncertainty propagation for A*B with B uncertain (helper function)
        n = len(A)
        RA = A[: n // 2]
        IA = A[n // 2 :]
        if isinstance(UB, float):  # simpler calculation if only single uncertainty
            uRR = RA * UB * RA + IA * UB * IA
            uRI = RA * UB * IA - IA * UB * RA
            uII = IA * UB * IA + RA * UB * RA
        elif len(UB.shape) == 1:  # simpler calculation if no correlation
            UBRR = UB[: n // 2]
            UBII = UB[n // 2 :]
            uRR = RA * UBRR * RA + IA * UBII * IA
            uRI = RA * UBRR * IA - IA * UBII * RA
            uII = IA * UBRR * IA + RA * UBII * RA
        else:  # full calculation because of full input covariance
            UBRR = UB[: n // 2, : n // 2]
            UBRI = UB[: n // 2, n // 2 :]
            UBII = UB[n // 2 :, n // 2 :]
            uRR = (
                _prod(RA, _prod(UBRR, RA))
                - _prod(IA, _prod(UBRI.T, RA))
                - _prod(RA, _prod(UBRI, IA))
                + _prod(IA, _prod(UBII, IA))
            )
            uRI = (
                _prod(RA, _prod(UBRR, IA))
                - _prod(IA, _prod(UBRI.T, IA))
                + _prod(RA, _prod(UBRI, RA))
                - _prod(IA, _prod(UBII, RA))
            )
            uII = (
                _prod(IA, _prod(UBRR, IA))
                + _prod(RA, _prod(UBRI.T, IA))
                + _prod(IA, _prod(UBRI, RA))
                + _prod(RA, _prod(UBII, RA))
            )
        return uRR, uRI, uII

    N = len(Y)
    RY = Y[: N // 2]
    IY = Y[N // 2 :]  # decompose into block matrix
    RF = F[: N // 2]
    IF = F[N // 2 :]  # decompose into block matrix
    YF = np.r_[RY * RF - IY * IF, RY * IF + IY * RF]  # apply product rule
    if not isinstance(UF, np.ndarray):  # second factor is known exactly
        UYRR, UYRI, UYII = _calcU(F, UY)
        # Stack together covariance matrix
        UYF = np.vstack((np.hstack((UYRR, UYRI)), np.hstack((UYRI.T, UYII))))
    else:  # both factors are uncertain
        URR_Y, URI_Y, UII_Y = _calcU(F, UY)
        URR_F, URI_F, UII_F = _calcU(Y, UF)
        URR = URR_Y + URR_F
        URI = URI_Y + URI_F
        UII = UII_Y + UII_F
        # Stack together covariance matrix
        UYF = np.vstack((np.hstack((URR, URI)), np.hstack((URI.T, UII))))
    return YF, UYF



# -*- coding: utf-8 -*-
"""

This example demonstrates the use of the transformation from real-imag representation to an amplitude-phase representation and back.
It uses the corresponding PyDynamic functions from the module `uncertainty.propagate_DFT´.

The function Time2AmpPhase uses GUM_DFT and DFT2AmpPhase internally.
The function AmpPhase2Time applies the formulas as given in the corresponding publication on GUM2DFT.

"""

import numpy as np
import matplotlib.pyplot as plt
from PyDynamic.uncertainty.propagate_DFT import AmpPhase2Time, Time2AmpPhase, GUM_DFTfreq, Time2AmpPhase_multi
from PyDynamic.misc.testsignals import multi_sine

# set amplitude values of multi-sine componentens (last parameter is number of components)
sine_amps = np.random.randint(1,4,10)
# set frequencies of multi-sine components
sine_freqs= np.linspace(100, 500, len(sine_amps))*2*np.pi
# define time axis
dt = 0.0001
time = np.arange(0.0, 0.2, dt)
# measurement noise standard deviation (assume white noise)
sigma_noise = 0.001
# generate test signal
testsignal = multi_sine(time, sine_amps, sine_freqs, noise=sigma_noise)

plt.figure(1, figsize=(12,6))
plt.plot(time, testsignal)
plt.xlabel("time in s")
plt.ylabel("signal amplitude in a.u.")

# uncertainty propagation from time domain to frequency domain
A,P, UAP = Time2AmpPhase(testsignal, sigma_noise**2)
f = GUM_DFTfreq(len(time), dt)

plt.figure(2, figsize=(12,6))
plt.errorbar(f, A, np.sqrt(np.diag(UAP)[:len(A)]), fmt=".-")
plt.xlabel("frequency in Hz")
plt.ylabel("DFT magnitude values in a.u.")

# uncertainty propagation from frequency domain to time domain
x, ux = AmpPhase2Time(A, P, UAP)

plt.figure(3, figsize=(12,6))
plt.subplot(211)
plt.errorbar(time, x, np.sqrt(np.diag(ux)), fmt=".-", label="after DFT and iDFT using amplitude and phase")
plt.plot(time, testsignal, label="original signal")
plt.xlabel("time in s")
plt.ylabel("signal amplitude in a.u.")
plt.legend()
plt.subplot(212)
plt.plot(time, np.sqrt(np.diag(ux)), label="uncertainty after DFT and iDFT using amplitude and phase")
plt.plot(time, np.ones_like(time)*sigma_noise, label="original uncertainty")
plt.xlabel("time in s")
plt.ylabel("uncertainty in a.u.")
plt.legend()

# # apply the same method to several signals with one call
M = 10
testsignals = np.zeros((M, len(time)))
for m in range(M):
    testsignals[m,:] = multi_sine(time, sine_amps, sine_freqs, noise = sigma_noise)

# select those frequencies of the 10 % largest magnitude values
indices = np.argsort(A)[:len(A)//10]
# propagate from time to frequency domain and only select the specific frequencies
A_mult, P_mult, UAP_mult = Time2AmpPhase_multi(testsignals, np.ones(M)*sigma_noise**2, selector=indices)

plt.figure(4, figsize=(12,6))
plt.subplot(211)
plt.plot(f[indices], A_mult.T, linestyle=":")
plt.subplot(212)
plt.plot(f[indices], P_mult.T, linestyle=":")


# -*- coding: utf-8 -*-
"""
The :mod:`PyDynamic.uncertainty` module implements methods for the
propagation of uncertainty for the most common use cases in the analysis of
dynamic measurements including filtering, Monte Carlo methods and time
series interpolation.

# See http://mathmet.org/projects/14SIP08 and
# https://www.github.com/eichstaedtPTB/PyDynamic
"""

from .propagate_DFT import (
    GUM_DFT,
    GUM_iDFT,
    DFT_deconv,
    DFT_multiply,
    DFT2AmpPhase,
    AmpPhase2DFT,
    AmpPhase2Time,
    Time2AmpPhase,
)

from .propagate_filter import FIRuncFilter, IIRuncFilter

from .propagate_MonteCarlo import MC, SMC, UMC, UMC_generic

from .interpolation import interp1d_unc

__all__ = [
    "GUM_DFT",
    "GUM_iDFT",
    "DFT_deconv",
    "DFT_multiply",
    "DFT2AmpPhase",
    "AmpPhase2DFT",
    "AmpPhase2Time",
    "Time2AmpPhase",
    "FIRuncFilter",
    "IIRuncFilter",
    "MC",
    "SMC",
    "UMC",
    "UMC_generic",
    "interp1d_unc",
]



"""Test PyDynamic.uncertainty.propagate_convolve"""

import numpy as np
import pytest
import scipy.ndimage as sn
from hypothesis import assume, given, strategies as hst

from PyDynamic.uncertainty.propagate_convolution import convolve_unc


def random_array(length):
    array = np.random.randn(length)
    return array


def random_covariance_matrix(length):
    """construct a valid (but random) covariance matrix with good condition number"""

    # because np.cov estimates the mean from data, the returned covariance matrix
    # has one eigenvalue close to numerical zero (rank n-1).
    # This leads to a singular matrix, which is badly suited to be used as valid
    # covariance matrix. To circumvent this:

    ## draw random (n+1, n+1) matrix
    cov = np.cov(np.random.random((length + 1, length + 1)))

    ## calculate SVD
    u, s, vh = np.linalg.svd(cov, full_matrices=False, hermitian=True)

    ## reassemble a covariance of size (n, n) by discarding the smallest singular value
    cov_adjusted = (u[:-1, :-1] * s[:-1]) @ vh[:-1, :-1]

    return cov_adjusted


def valid_inputs(reduced_set=False):

    valid_inputs = []

    for n in [10, 15, 20]:
        x_signal = random_array(n)
        u_signal = random_covariance_matrix(n)

        if reduced_set:
            valid_inputs.append([x_signal, u_signal])
        else:
            valid_inputs.append([x_signal, None])
            valid_inputs.append([x_signal, np.diag(np.diag(u_signal))])
            valid_inputs.append([x_signal, u_signal])

    return valid_inputs


def valid_modes(kind="all"):
    scipy_modes = {"nearest", "reflect", "mirror"}
    numpy_modes = {"full", "valid", "same"}

    if kind == "all":
        return numpy_modes.union(scipy_modes)
    elif kind == "scipy":
        return scipy_modes
    elif kind == "numpy":
        return numpy_modes
    else:
        return set()


@pytest.mark.parametrize("input_1", valid_inputs())
@pytest.mark.parametrize("input_2", valid_inputs())
@pytest.mark.parametrize("mode", valid_modes())
def test_convolution(input_1, input_2, mode):

    scipy_modes = valid_modes("scipy")
    numpy_modes = valid_modes("numpy")

    # calculate the convolution of x1 and x2
    y, Uy = convolve_unc(*input_1, *input_2, mode)

    if mode in numpy_modes:
        y_ref = np.convolve(input_1[0], input_2[0], mode=mode)
    elif mode in scipy_modes:
        y_ref = sn.convolve1d(input_1[0], input_2[0], mode=mode)

    # compare results
    assert len(y) == len(Uy)
    assert len(y) == len(y_ref)
    assert np.allclose(y, y_ref)


@pytest.mark.parametrize("input_1", valid_inputs(reduced_set=True))
@pytest.mark.parametrize("input_2", valid_inputs(reduced_set=True))
@pytest.mark.parametrize("mode", valid_modes())
def test_convolution_monte_carlo(input_1, input_2, mode):

    scipy_modes = valid_modes("scipy")
    numpy_modes = valid_modes("numpy")

    # pydynamic calculation
    y, Uy = convolve_unc(*input_1, *input_2, mode)

    # Monte Carlo simulation
    mc_results = []
    n_runs = 20000
    XX1 = np.random.multivariate_normal(*input_1, size=n_runs)
    XX2 = np.random.multivariate_normal(*input_2, size=n_runs)
    for x1, x2 in zip(XX1, XX2):
        if mode in numpy_modes:
            conv = np.convolve(x1, x2, mode=mode)
        elif mode in scipy_modes:
            conv = sn.convolve1d(x1, x2, mode=mode)
        mc_results.append(conv)

    y_mc = np.mean(mc_results, axis=0)
    Uy_mc = np.cov(mc_results, rowvar=False)

    # HACK: for visualization during debugging
    # import matplotlib.pyplot as plt
    # fig, ax = plt.subplots(nrows=1, ncols=4)
    # _min = min(Uy.min(), Uy_mc.min())
    # _max = max(Uy.max(), Uy_mc.max())
    # ax[0].plot(y, label="fir")
    # ax[0].plot(y_mc, label="mc")
    # ax[0].set_title("mode: {0}, x1: {1}, x2: {2}".format(mode, len(x1), len(x2)))
    # ax[0].legend()
    # ax[1].imshow(Uy, vmin=_min, vmax=_max)
    # ax[1].set_title("PyDynamic")
    # ax[2].imshow(Uy_mc, vmin=_min, vmax=_max)
    # ax[2].set_title("numpy MC")
    # img = ax[3].imshow(np.log(np.abs(Uy-Uy_mc)))
    # ax[3].set_title("log(abs(diff))")
    # fig.colorbar(img, ax=ax[3])
    # plt.show()
    # /HACK

    assert np.allclose(y, y_mc, rtol=1e-1, atol=1e-1)
    assert np.allclose(Uy, Uy_mc, rtol=1e-1, atol=1e-1)


@pytest.mark.parametrize("input_1", valid_inputs(reduced_set=True))
@pytest.mark.parametrize("input_2", valid_inputs(reduced_set=True))
@given(hst.text())
def test_convolution_invalid_mode(input_1, input_2, mode):
    set_of_valid_modes = set(valid_modes())
    assume(mode not in set_of_valid_modes)

    with pytest.raises(ValueError):
        convolve_unc(*input_1, *input_2, mode)



"""
The convolution operation is a common operation in signal and data
processing. Convolving signals is mathematically similar to a filter
application. 

This module contains the following function:

* :func:`convolve_unc`: Convolution with uncertainty propagation based on FIR-filter

"""

import numpy as np

from .propagate_filter import _fir_filter

__all__ = ["convolve_unc"]


def convolve_unc(x1, U1, x2, U2, mode="full"):
    """
    An implementation of the discrete convolution of two signals with uncertainty propagation.
    It supports the convolution modes of :func:`numpy.convolve` and :func:`scipy.ndimage.convolve1d`.

    Parameters
    ----------
    x1 : np.ndarray, (N,)
        first input signal
    U1 : np.ndarray, (N, N)
        full 2D-covariance matrix associated with x1
        if the signal is fully certain, use ``U1 = None`` to make use of more efficient calculations.
    x2 : np.ndarray, (M,)
        second input signal
    U2 : np.ndarray, (M, M)
        full 2D-covariance matrix associated with x2
        if the signal is fully certain, use ``U2 = None`` to make use of more efficient calculations.
    mode : str, optional
        :func:`numpy.convolve`-modes:

        - full:  len(y) == N+M-1 (default)
        - valid: len(y) == max(M, N) - min(M, N) + 1
        - same:  len(y) == max(M, N) (value+covariance are padded with zeros)

        :func:`scipy.ndimage.convolve1d`-modes:

        - nearest: len(y) == N (value+covariance are padded with by stationary assumption)
        - reflect:  len(y) == N
        - mirror:   len(y) == N

    Returns
    -------
    conv : np.ndarray
        convoluted output signal
    Uconv : np.ndarray
        full 2D-covariance matrix of y

    References
    ----------
    .. seealso::
        :func:`numpy.convolve`
        :func:`scipy.ndimage.convolve1d`
    """

    # if a numpy-mode is chosen, x1 is expected to be the longer signal
    # remember that pure convolution is commutative
    if len(x1) < len(x2) and mode in ["valid", "full", "same"]:
        x1, x2 = x2, x1
        U1, U2 = U2, U1

    # actual computation
    if mode == "valid":
        # apply _fir_filter directly
        y, Uy = _fir_filter(x=x1, theta=x2, Ux=U1, Utheta=U2, initial_conditions="zero")

        # compensate boundary adjustments from _fir_filter
        conv = y[len(x2) - 1 :]
        Uconv = Uy[len(x2) - 1 :, len(x2) - 1 :]

    elif mode == "full":
        # append zeros to adapt to _fir_filter mechanism
        pad_len = len(x2) - 1
        x1_mod = np.pad(x1, (0, pad_len), mode="constant", constant_values=0)
        if isinstance(U1, np.ndarray):
            U1_mod = np.pad(
                U1, ((0, pad_len), (0, pad_len)), mode="constant", constant_values=0
            )
        else:
            U1_mod = None

        # apply _fir_filter
        y, Uy = _fir_filter(
            x=x1_mod, theta=x2, Ux=U1_mod, Utheta=U2, initial_conditions="zero"
        )

        # use output directly
        conv = y
        Uconv = Uy

    elif mode == "same":
        # append zeros to adapt to _fir_filter mechanism
        pad_len = (len(x2) - 1) // 2
        x1_mod = np.pad(x1, (0, pad_len), mode="constant", constant_values=0)
        if isinstance(U1, np.ndarray):
            U1_mod = np.pad(
                U1, ((0, pad_len), (0, pad_len)), mode="constant", constant_values=0
            )
        else:
            U1_mod = None

        # apply _fir_filter
        y, Uy = _fir_filter(
            x=x1_mod, theta=x2, Ux=U1_mod, Utheta=U2, initial_conditions="zero"
        )

        # compensate boundary adjustments from _fir_filter
        conv = y[pad_len:]
        Uconv = Uy[pad_len:, pad_len:]

    elif mode in ["nearest", "reflect", "mirror"]:

        # scipy.ndimage.convolve1d and numpy.pad use different (but overlapping) terminology
        mode_translation = {
            "nearest": "edge",
            "reflect": "symmetric",
            "mirror": "reflect",
        }
        pad_mode = mode_translation[mode]

        # prepend and append to x1 and U1 to get requested boundary effect
        n1 = len(x1)
        n2 = len(x2)
        pad_len = (n2 + 1) // 2
        x1_mod = np.pad(x1, (pad_len, pad_len), mode=pad_mode)
        # we assume that U1 is an array or None
        if isinstance(U1, np.ndarray):
            U1_mod = np.pad(U1, ((pad_len, pad_len), (pad_len, pad_len)), mode=pad_mode)
        else:
            U1_mod = None

        # apply _fir_filter
        y, Uy = _fir_filter(
            x=x1_mod, theta=x2, Ux=U1_mod, Utheta=U2, initial_conditions="zero"
        )

        # compensate boundary adjustments from _fir_filter
        conv = y[n2 : n2 + n1]
        Uconv = Uy[n2 : n2 + n1, n2 : n2 + n1]

    else:
        raise ValueError(f'convolve_unc: Mode "{mode}" is not supported.')

    return conv, Uconv
